!
!==========================================================================================
!
! ht.f90:               Optimal Heavy Tail Estimation
!                       =============================
!
!                       Heavy tail index estimation (Hill or Pickands estimator)
!                       on time series data by means of a brute force order selector
!                       (RMSE measure), including preprocessing routines.
!
!==========================================================================================
!
! Copyright notice:     Copyright (C) Climate Risk Analysis - Manfred Mudelsee
!                       http://www.climate-risk-analysis.com
!
! Permission notice:    Permission is hereby granted, free of charge, to any person
!                       obtaining a copy of this software and associated computer files---
!                       excluding the Numerical Recipes file---(the "Software"), to
!                       deal in the Software without restriction, including without
!                       limitation the rights to use, copy, modify, merge, publish,
!                       distribute, sublicense, and/or sell copies of the Software, and to
!                       permit persons to whom the Software is furnished to do so, subject
!                       to the following conditions:
!
!                       The above copyright notice, including the URL, and this permission
!                       notice shall be included in all copies or substantial portions of
!                       the Software.
!
!==========================================================================================
!
! Disclaimer:           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
!                       EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
!                       MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
!                       NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
!                       BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
!                       ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
!                       CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
!                       SOFTWARE.
!
!==========================================================================================
!
! Developer:            Climate Risk Analysis - Manfred Mudelsee
!                       Kreuzstrasse 27
!                       Heckenbeck
!                       37581 Bad Gandersheim
!                       Germany
!
! Contact:              http://www.climate-risk-analysis.com
!                       mudelsee@climate-risk-analysis.com
!
!==========================================================================================
!
! Language:             Fortran 90
!
!==========================================================================================
!
! Development systems:  o IBM Intelli Workstation Z Pro
!                       o 4 CPU Intel Xeon 5160 (8 GB RAM) (allows for 4 threads)
!                       o Windows XP Professional x64 Edition Service Pack 2
!
!                       o Tarox Workstation 745TQ
!                       o 2 CPU Intel Xeon E5-2620 (64 GB RAM) (allows for 24 threads)
!                       o Linux Fedora 15
!
!==========================================================================================
!
! Compiler:             Intel(R) Visual Fortran Composer XE 2013
!
! Compiling options:    final development:
!                        /nologo /O2 /Qipo /Qopenmp /Qdiag-disable:8290
!                        /module:"Release\Modules" /object:"Release\Obj\ht"
!                        /Fd"Release\vc100.pdb" /traceback /check:bounds
!                        /libs:static /threads /c
!
! Linker hints:         Large data sizes may lead to stack overflows, notably in a
!                       parallel computing implementation. In such a case, it is helpful
!                       to increase the computer's stack and heap sizes; this can be done
!                       at the step of linking; the suitable values have to be found in
!                       an explorative manner.
!
!                       Example: for a data size of 150,000 on the IBM development  
!                       (see above), suitable stack-heap parameters were:
!                       heap  reserve size: 100,000,000
!                       heap   commit size:   1,000,000
!                       stack reserve size: 600,000,000
!                       stack  commit size:   6,000,000
!
!==========================================================================================
!
! Change log:
!
!                       10/2017             replacement of subroutine stable
!                                           by subroutine expn01 for order selection
!
!                       05/2017             manual (Version 1.0) written
!
!        Version 1.0    05/2017
!
!                       04/2017-05/2017     tests (number of MC simulations), accuracy
!
!                       03/2017             random number generator seed moved from
!                                           subroutines generate and generateMC to MAIN
!
!                       02/2017             RMSE calculation changed to MC
!
!                       12/2016-01/2017     automation via configuration files
!
!                       05/2016-09/2016     tests on artificial data,
!                                           addition of preprocessing options,
!                                           work on memory usage optimization
!
!                       01/2015             first cleaning and commenting,
!                                           preprocessing routine 'time reversal'
!
!                       07/2014             transfer to final development, initial tests
!                                           (Manfred Mudelsee)
!
!                       09/2012-05/2014     original development
!                                           (Miguel Bermejo, Manfred Mudelsee)
!
!==========================================================================================
!
! Funding:              This work is supported by the European Commission via the
!                       Marie Curie Initial Training Network LINC (project number 289447)
!                       under the Seventh Framework Programme.
!
! Reference:            Mudelsee M, Bermejo MA: Optimal heavy tail estimation,
!                                               Part I: Order selection.
!                                               Nonlinear Processes in Geophysics,
!                                               VOL, URL, 2017.
!
!==========================================================================================
!
! Additional software:  An output analysis software (ht-d-stats) calculates descriptive
!                       statistical properties of input data and generates plot data.
!
!                       The source, ht-d-stats.f90, has to be self compiled; some
!                       copyrighted Numerical Recipes source files are additionaly required
!                       for this. The executable (ht-d-stats.exe) has to be placed in a
!                       subdirectory (of arbitrary name) of ht.exe.
!
!                       The files ht-d-stats.f90 and ht-d-stats.exe (a Windows version)
!                       are available upon request from the developer.
!
!==========================================================================================
!
! Files to be included
! ====================
!
include 'nrtype.f90'    ! Numerical Recipes, free software, modified to 'double precision'
include 'rng.f90'       ! random number generator for parallel computing, requires file
!                       ! 'mzranseed.dat' to reside in the same directory as ht.exe,
                        ! see http://www.manfredmudelsee.com/soft/rng,
                        ! the rng.f90 version to be used with ht.f90 is modified in one
                        ! point: parameter ns1 is public (for automation, modus 'generate')
!
module version
      implicit none
      character(len=18), parameter :: versiondate='2.0 (October 2017)'
end module version
!
module cfgseedparameters
      use nrtype
      implicit none
      character(len=8)      :: modus                            = '--------'
      character(len=1000)   :: SUBR_READ_FILE_inputfile         = '--------'
      integer               :: MAIN_opt2                        = -999
      integer               :: MAIN_optBF                       = -999
      integer               :: MAIN_L_k                         = -999
      integer               :: MAIN_Pmink                       = -999
      character(len=1)      :: MAIN_mctauflag                   = '-'
      character(len=80)     :: SUBR_EST_OUT_outputfile1         = 'htresult.dat'
      character(len=80)     :: SUBR_EST_OUT_outputfile2         = '--------'
      integer               :: MAIN_Norig                       = -999
      integer               :: MAIN_nsim                        = -999
      real(sp)              :: SUBR_GENERATE_tau                = -999.0_sp
      real(sp)              :: SUBR_GENERATE_alpha              = -999.0_sp
      character(len=1)      :: SUBR_GENERATE_shape              = '-'
      character(len=1000)   :: SUBR_GENERATE_artfile            = '--------'
      integer               :: htiseed                          = -999
      integer               :: htiseed2                         = -999
end module cfgseedparameters

module datafilename
      implicit none
      character(len=1000) :: inputfile='[NOT YET DEFINED]'  ! input data file name
end module datafilename

module Hill_Pick_testNpos_alpha
      use nrtype
      integer :: itestnp=-999                               ! counter, Hill estimator, test for Npositive .ge. minNpositive
      integer, parameter :: ntestnp = 100                   ! upper bound of itestnp
      integer :: testNpositive=-999                         ! tested Npositive
      integer, parameter :: minNpositive = 2                ! required minimum of Npositive
      real(sp), parameter :: fracp=0.20_sp                  ! Pickands estimator, N4 = INT(Nobs*fracp)
                                                            ! fracp must be .le. 0.25 (see subroutine pickands)
end module Hill_Pick_testNpos_alpha

module MC
      use nrtype
      implicit none
      integer :: dummy_MC_k_opt=-999                        ! MC: dummy
      integer :: dummy_MC_K_work=-999                       ! MC: dummy
      real(sp) :: dummy_tau=-999.0_sp                       ! MC: dummy
      real(sp), dimension(:), allocatable :: Ysim           ! MC: simulated variable
      real(sp), dimension(:), allocatable :: tau_est_sim    ! MC: simulated estimated persistence time
      real(sp), dimension(:), allocatable :: alpha_est_sim  ! MC: simulated estimated tail index (alpha)
      real(sp), dimension(:), allocatable :: gamma_est_sim  ! MC: simulated estimated tail index (gamma)
end module MC

module qbfparameters
      implicit none
      integer :: optBF      ! optBF = 1: brute force
                            ! optBF = 2: quasi-brute force
      integer :: L_k=5      ! Quasi-brute force, Step 1: length of search-steps
                            ! L_k=5 may be a good default for data sizes not too large;
                            ! larger data sizes may require larger L_k
      integer :: Pmink=5    ! Quasi-brute force, Step 2: percentage of best k values
                            ! from Step 1 (in terms of RMSE), see subroutine est_out
                            ! Pmink=5 may be a good default for data sizes not too large;
                            ! larger data sizes may require smaller Pmink
                            !
                            ! The formula is: if
                            ! CPU (brute force) ~ n, then
                            ! CPU (quasi-brute force) ~ n * (1/L_k + Pmink/100)
end module qbfparameters

module result1
      use nrtype
      implicit none
      real(sp) :: tau_est_result        = -999.0_sp                 ! result: estimated persistence time
      real(sp) :: tau_est_RMSE_result   = -999.0_sp                 ! result: RMSE(tau_est_result)
      integer  :: k_opt_result          = -999                      ! result: optimal order
      integer  :: K_result              = -999                      ! result: maximum possible search value for the order
      real(sp) :: alpha_est_result      = -999.0_sp                 ! result: estimated tail index (alpha)
      real(sp) :: gamma_est_result      = -999.0_sp                 ! result: estimated tail index (gamma)
      real(sp) :: alpha_est_RMSE_result = -999.0_sp                 ! result: RMSE(alpha_est_result)
      real(sp) :: gamma_est_RMSE_result = -999.0_sp                 ! result: RMSE(gamma_est_result)
      real(sp), dimension(:), allocatable :: alpha_sequence_result  ! result: sequence of tail index estimates for all possible k values
      real(sp), dimension(:), allocatable :: RMSE_result            ! result: RMSE measure for order selection
end module result1

program menu

    !--------
    ! MODULES
    !--------

    use nrtype
    use rng
    use omp_lib
    use cfgseedparameters
    use datafilename
    use Hill_Pick_testNpos_alpha
    use MC
    use qbfparameters
    use result1
    use version
    implicit none

    !-----------
    ! PARAMETERS
    !-----------

    character(len=1), parameter :: pickg = 'n'  ! pickg = 'y': Pickands estimator calculates gamma = 1/alpha;
                                                !              should be numerically advantageous;
                                                !              note that source code variables named 'alpha'
                                                !              (e.g., alpha_est) may sometimes actually denote gamma
                                                !              (exception: alpha_sim)
                                                ! pickg = 'n': Pickands estimator calculates alpha
                                                ! 
    integer, parameter :: N_inner = 1000        ! number of internal loops (brute force order selector)
    ! integer, parameter :: Nobsmax = 250000      ! maximum number of observations
    integer, parameter :: Nobsmax = 10000000      ! maximum number of observations
    integer, parameter :: Nobsmin = 10          ! minimum number of observations

    !----------
    ! VARIABLES
    !----------

    ! Control variables for MENU
    integer :: opt, opt2

    ! Data
    real(sp), dimension(:), allocatable     :: t          ! time (possibly preprocessed)
    real(sp), dimension(:), allocatable     :: Y          ! variable (possibly preprocessed)
    integer                                 :: Nobs       ! data size (t, Y)
    real(sp), dimension(:), allocatable     :: ttrans     ! time (temporarily used during preprocessing for transformation)
    real(sp), dimension(:), allocatable     :: Ytrans     ! variable (temporarily used during preprocessing for transformation)
    integer                                 :: Ntrans     ! data size (ttrans, Ytrans)
    real(sp), dimension(:), allocatable     :: torig      ! time (original)
    real(sp), dimension(:), allocatable     :: Yorig      ! variable (original)
    integer                                 :: Norig      ! data size (torig, Yorig)

    ! Counters
    integer :: isim

    ! Auxiliary variables and parameters
    character                    :: answer
    integer                      :: col
    integer                      :: iocheck=-999
    logical                      :: dataloaded

    ! Random number generator
    character(len=10), parameter :: htseedfile='htseed.dat'
    type(rng_t), dimension(1), save :: rng3               ! seed of random number generator
    type(rng_t), dimension(1), save :: rng4               ! seed of random number generator
    integer :: iseed=1                                    ! dummy integer for random number generator

    ! Command line argument
    character(len=1000) :: arg  ! configuration file name
    
    ! Flag for estimation in subroutine est_out
    character(len=1) :: estflag='-'
    
    ! Flag for output in subroutine est_out
    character(len=1) :: outflag='-'
    
    ! Flags for persistence time estimation in subroutine est_out
    character(len=1) :: tauflag='-'
    character(len=1) :: mctauflag='-'
    
    ! Flags for testing whether alpha is between 0 and 2.
    character(len=1) :: flagt='-'
    character(len=1) :: flagtMC='-'
    
    ! Number of Monte Carlo (MC) simulations for RMSE calculations
    integer :: nsim=-999

    ! Number of threads
    integer :: nthreads

    ! Configuration file variables
    namelist /cfg/ modus, SUBR_READ_FILE_inputfile, MAIN_opt2, MAIN_optBF, MAIN_L_k, MAIN_Pmink, MAIN_mctauflag,                        &
                   SUBR_EST_OUT_outputfile1, SUBR_EST_OUT_outputfile2, MAIN_Norig, MAIN_nsim,                                           &
                   SUBR_GENERATE_tau, SUBR_GENERATE_alpha, SUBR_GENERATE_shape, SUBR_GENERATE_artfile

    !------------
    ! PRELIMINARY
    !------------

    ! Configuration file via command line
    call get_command_argument(1,arg)
    if (trim(adjustl(arg)) .eq. '') then
       modus='per-hand'
    else
       open (unit=1, file=trim(adjustl(arg)), status='old', form='formatted', action='read')
       read (unit=1, nml = cfg)
       close (unit=1, status='keep')
    end if

    ! Starting auxiliary
    dataloaded=.false.

    ! Starting the parallel random number generator
    call rng_initial_seed

    ! Seed (modus = 'per-hand' or modus = 'generate')
    ! The purpose is to secure independent realizations in case ht.exe is called several times
    ! (e.g., for 'external' MC simulations).
    ! Note that subroutine generate uses two copies (rng3, rng4), hence two seeds.
    open (unit=1, file=htseedfile, status='old', form='formatted', action='readwrite', iostat = iocheck)
    if (iocheck .eq. 0 ) then
       read (unit=1, fmt=*) htiseed
       htiseed2=mod(htiseed,ns1)+1
       rewind(1)
       write (unit=1, fmt='(i6)') mod(htiseed+1,ns1)+1
       close (unit=1, status='keep')
    else
       ! if no seedfile exists, start a new one
       htiseed=1
       htiseed2=2
       open (unit=1, file=htseedfile, status='new', form='formatted', action='write')
       write (unit=1, fmt='(i6)') htiseed+2
       close (unit=1, status='keep')
    end if

    ! Seed random number generator
    call rng_seed(rng3(iseed),htiseed)
    call rng_seed(rng4(iseed),htiseed2)

    ! Setting the parallelization
    nthreads=omp_get_max_threads()
    call omp_set_num_threads(nthreads)

    ! Welcome
    call clearscreen
    print *
    print '(a)',   ' Optimal Heavy Tail Estimation'
    print '(a)',   ' ============================='
    print *
    print '(a)',   ' Heavy tail index estimation (Hill or Pickands estimator)'
    print '(a)',   ' on time series data by means of a brute force order selector'
    print '(a)',   ' (RMSE measure), including preprocessing routines.'
    print *
    print '(a,a)', ' Version ',versiondate
    print *
    print '(a)',   ' Copyright (C) Climate Risk Analysis - Manfred Mudelsee'
    print '(a)',   ' http://www.climate-risk-analysis.com'
    print *
    print '(a)',   ' Please see source code (ht.f90) for disclaimer and permissions.'
    print *
    print '(a)',   ' PRESS ENTER TO CONTINUE'
    if (modus .eq. 'per-hand') read (*,*)
    call clearscreen
    print *
    if (nthreads .lt. 10) then
        print '(a,i1)', ' Parallel computing, determined number of threads: ',nthreads
    else if (nthreads .ge. 10    .and. nthreads .lt. 100)    then
        print '(a,i2)', ' Parallel computing, determined number of threads: ',nthreads
    else if (nthreads .ge. 100   .and. nthreads .lt. 1000)   then
        print '(a,i3)', ' Parallel computing, determined number of threads: ',nthreads
    else if (nthreads .ge. 1000  .and. nthreads .lt. 10000)  then
        print '(a,i4)', ' Parallel computing, determined number of threads: ',nthreads
    else if (nthreads .ge. 10000 .and. nthreads .lt. 100000) then
        print '(a,i5)', ' Parallel computing, determined number of threads: ',nthreads
    else if (nthreads .ge. 100000) then
        print '(a,i6)', ' Parallel computing, determined number of threads: ',nthreads
    end if
    print *

    !-------------
    ! MAIN PROGRAM
    !-------------

    if (modus .eq. 'per-hand') then
       do
           opt=choices_main(dataloaded)
           select case (opt)
               case (1)                                                 ! data file
                   call read_file(t,Y,Nobs,torig,Yorig,Norig,inputfile)
               case (2)                                                 ! preprocessing
                   do
                       opt2=menu_data()
                       select case (opt2)
                           case (1)                                     ! reverse time direction
                                t=t(Nobs:1:-1)
                                Y=Y(Nobs:1:-1)
                           case (2)                                     ! transform the data otherwise
                                Ntrans=Nobs
                                allocate(ttrans(1:Ntrans))
                                allocate(Ytrans(1:Ntrans))
                                ttrans=t
                                Ytrans=Y
                                call menu_transform(Y,t,Nobs,Ytrans,ttrans,Ntrans)
                                if (Ntrans .eq. Nobs) then              ! unchanged data size
                                   t=ttrans
                                   Y=Ytrans
                                else                                    ! changed data size (e.g., 1st differences)
                                   deallocate(t)
                                   deallocate(Y)
                                   allocate(t(1:Ntrans))
                                   allocate(Y(1:Ntrans))
                                   t(1:Ntrans)=ttrans(1:Ntrans)
                                   Y(1:Ntrans)=Ytrans(1:Ntrans)
                                end if
                                Nobs=Ntrans
                                deallocate(ttrans)
                                deallocate(Ytrans)
                           case default
                                exit
                       end select
                   end do
               case (3)                                                 ! tail index estimation
                   do
                       opt2=menu_estimator()
                       if (opt2 .eq. 3) then
                           exit
                       else
                           call menu_order_selection(Nobs)              ! sets optBF (and eventually L_k, Pmink)
                           nsim=menu_nsim()
                           if (nsim .eq. -999) then
                              tau_est_RMSE_result=-999.0_sp
                              alpha_est_RMSE_result=-999.0_sp
                              gamma_est_RMSE_result=-999.0_sp
                              ! Test (over the full order search range) whether at least for one order value
                              ! the tail index alpha is between 0 and 2.
                              call test_alpha_initial(Y,Nobs,opt2,flagt)
                              if (flagt .eq. '+') then
                                 estflag='y'                            ! estimate without RMSE and with output
                                 outflag='y'
                                 tauflag='y'
                                 call    est_out(estflag,outflag,  tauflag,dummy_tau     ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                                                 inputfile,         k_opt_result,         K_result,nsim,                           &
                                                 tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                                                 alpha_est_RMSE_result,gamma_est_RMSE_result)
                                                 !
                                                 ! note: subroutine est_out (under estflag='y') allocates (in case of not allocated)
                                                 !       alpha_sequence_result, RMSE_result
                                                 !
                                 if (k_opt_result > 0) call writeykopt(t,Y,Nobs,k_opt_result) ! output analysis software ht-d-stats
                                 if (allocated(alpha_sequence_result)) deallocate(alpha_sequence_result)
                                 if (allocated(RMSE_result))           deallocate(RMSE_result)
                              else if (flagt .eq. '-') then
                                 estflag='n'                            ! no estimation possible
                                 outflag='y'
                                 call    est_out(estflag,outflag,  tauflag,dummy_tau     ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                                                 inputfile,         k_opt_result,         K_result,nsim,                           &
                                                 tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                                                 alpha_est_RMSE_result,gamma_est_RMSE_result)
                                                 !
                                                 ! note: subroutine est_out (under estflag='y') allocates (in case of not allocated)
                                                 !       alpha_sequence_result, RMSE_result
                                                 !
                                 if (allocated(alpha_sequence_result)) deallocate(alpha_sequence_result)
                                 if (allocated(RMSE_result))           deallocate(RMSE_result)
                              end if
                           else
                              ! Test (over the full order search range) whether at least for one order value
                              ! the tail index alpha is between 0 and 2.
                              call test_alpha_initial(Y,Nobs,opt2,flagt)
                              if (flagt .eq. '+') then
                                 estflag='y'                            ! first, estimate without RMSE and without output
                                 outflag='n'
                                 tauflag='y'
                                 call menu_MC_tauflag(mctauflag)
                                 call    est_out(estflag,outflag,  tauflag,dummy_tau     ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                                                 inputfile,         k_opt_result,         K_result,nsim,                           &
                                                 tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                                                 alpha_est_RMSE_result,gamma_est_RMSE_result)
                                                 !
                                                 ! note: subroutine est_out (under estflag='y') allocates (in case of not allocated)
                                                 !       alpha_sequence_result, RMSE_result
                                                 !
                                 allocate(Ysim(Nobs))                   ! second, calculate RMSE values using MC
                                 Ysim=-999.0_sp
                                 allocate(tau_est_sim(nsim))
                                 tau_est_sim=-999.0_sp
                                 allocate(alpha_est_sim(nsim))
                                 alpha_est_sim=-999.0_sp
                                 allocate(gamma_est_sim(nsim))
                                 gamma_est_sim=-999.0_sp
                                 do isim=1,nsim
                                    print *
                                    print *, ' -----------------------------------------------------------'
                                    print *, ' Monte Carlo: simulation       ', isim,' of ', nsim
                                    print *, ' -----------------------------------------------------------'
testnp1:                            do itestnp=1,ntestnp
                                       ! at least minNpositive positive values of mean-subtracted Ysim are required for Hill estimator
                                       call generateMC(t,Ysim,Nobs,tau_est_result,alpha_est_result,gamma_est_result,opt2,pickg,rng3(iseed),rng4(iseed))
                                       testNpositive=count((Ysim-sum(Ysim)/Nobs) > 0.0_sp)
                                       ! Test (over the full order search range) whether at least for one order value
                                       ! the tail index alpha (on the simulated data) is between 0 and 2.
                                       call test_alpha_initial(Ysim,Nobs,opt2,flagtMC)
                                       if (flagtMC .eq. '+' .and. .not. (opt2 .eq. 1 .and. testNpositive .lt. minNpositive)) exit testnp1
                                    end do testnp1
                                    call est_out(estflag,outflag,mctauflag,tau_est_result,t,Ysim,torig,Yorig,Nobs,Norig,opt2,      &
                                                 inputfile,dummy_MC_k_opt       ,dummy_MC_K_work  ,nsim,                           &
                                                 tau_est_sim(isim),tau_est_RMSE_result,alpha_est_sim(isim),gamma_est_sim(isim),    &
                                                 alpha_est_RMSE_result,gamma_est_RMSE_result)
                                 end do
                                 if (mctauflag .eq. 'y')                                                                           &
                                    call calc_RMSE(  tau_est_sim,nsim,  tau_est_result,  tau_est_RMSE_result)
                                 if (opt2 .eq. 2 .and. pickg .eq. 'y') then
                                    call calc_RMSE(gamma_est_sim,nsim,gamma_est_result,gamma_est_RMSE_result)
                                 else
                                    call calc_RMSE(alpha_est_sim,nsim,alpha_est_result,alpha_est_RMSE_result)
                                 end if
                                 estflag='n'                            ! third, output estimates and RMSE
                                 outflag='y'
                                 call    est_out(estflag,outflag,  tauflag,tau_est_result,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                                                 inputfile,         k_opt_result,         K_result,nsim,                           &
                                                 tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                                                 alpha_est_RMSE_result,gamma_est_RMSE_result)
                                 if (k_opt_result > 0) call writeykopt(t,Y,Nobs,k_opt_result) ! output analysis software ht-d-stats
                                 deallocate(gamma_est_sim)
                                 deallocate(alpha_est_sim)
                                 deallocate(tau_est_sim)
                                 deallocate(Ysim)
                                 if (allocated(alpha_sequence_result)) deallocate(alpha_sequence_result)
                                 if (allocated(RMSE_result))           deallocate(RMSE_result)
                              else if (flagt .eq. '-') then
                                 estflag='n'                            ! no estimation possible
                                 outflag='y'
                                 call    est_out(estflag,outflag,  tauflag,dummy_tau     ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                                                 inputfile,         k_opt_result,         K_result,nsim,                           &
                                                 tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                                                 alpha_est_RMSE_result,gamma_est_RMSE_result)
                                                 !
                                                 ! note: subroutine est_out (under estflag='y') allocates (in case of not allocated)
                                                 !       alpha_sequence_result, RMSE_result
                                                 !
                                 if (allocated(alpha_sequence_result)) deallocate(alpha_sequence_result)
                                 if (allocated(RMSE_result))           deallocate(RMSE_result)
                              end if
                           end if
                       end if
                   end do
               case (4)                                                 ! artificial data generation
                   call generate(torig,Norig,rng3(iseed),rng4(iseed))
               case default                                             ! exit
                   exit
               end select
           call clearscreen
       end do
    else if (modus .eq. 'estimate') then
       opt2=MAIN_opt2
       optBF=MAIN_optBF
       if (optBF .eq. 2) then
          L_k=MAIN_L_k
          Pmink=MAIN_Pmink
       end if
       nsim=MAIN_nsim
       call read_file(t,Y,Nobs,torig,Yorig,Norig,inputfile)
       if (nsim .eq. -999) then
           ! Test (over the full order search range) whether at least for one order value
           ! the tail index alpha is between 0 and 2.
           call test_alpha_initial(Y,Nobs,opt2,flagt)
           if (flagt .eq. '+') then
              estflag='y'                   ! estimate without RMSE and with output
              outflag='y'
              tauflag='y'
              call    est_out(estflag,outflag,  tauflag,dummy_tau     ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,     &
                              inputfile,         k_opt_result,         K_result,nsim,                          &
                              tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,            &
                              alpha_est_RMSE_result,gamma_est_RMSE_result)
                              !
                              ! note: subroutine est_out (under estflag='y') allocates (in case of not allocated)
                              !       alpha_sequence_result, RMSE_result
                              !
              if (k_opt_result > 0) call writeykopt(t,Y,Nobs,k_opt_result) ! output analysis software ht-d-stats
              if (allocated(alpha_sequence_result)) deallocate(alpha_sequence_result)
              if (allocated(RMSE_result))           deallocate(RMSE_result)
           else if (flagt .eq. '-') then
              estflag='n'                   ! no estimation possible
              outflag='y'
              call    est_out(estflag,outflag,  tauflag,dummy_tau     ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                              inputfile,         k_opt_result,         K_result,nsim,                           &
                              tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                              alpha_est_RMSE_result,gamma_est_RMSE_result)
                              !
                              ! note: subroutine est_out (under estflag='y') allocates (in case of not allocated)
                              !       alpha_sequence_result, RMSE_result
                              !
              if (allocated(alpha_sequence_result)) deallocate(alpha_sequence_result)
              if (allocated(RMSE_result))           deallocate(RMSE_result)
           end if
       else
           ! Test (over the full order search range) whether at least for one order value
           ! the tail index alpha is between 0 and 2.
           call test_alpha_initial(Y,Nobs,opt2,flagt)
           if (flagt .eq. '+') then
              estflag='y'                       ! first, estimate without RMSE and without output
              outflag='n'
              tauflag='y'
              mctauflag=MAIN_mctauflag
              call    est_out(estflag,outflag,  tauflag,dummy_tau     ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                              inputfile,         k_opt_result,         K_result,nsim,                           &
                              tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                              alpha_est_RMSE_result,gamma_est_RMSE_result)
                              !
                              ! note: subroutine est_out (under estflag='y') allocates (in case of not allocated)
                              !       alpha_sequence_result, RMSE_result
                              !
              allocate(Ysim(Nobs))              ! second, calculate RMSE values using MC
              Ysim=-999.0_sp
              allocate(tau_est_sim(nsim))
              tau_est_sim=-999.0_sp
              allocate(alpha_est_sim(nsim))
              alpha_est_sim=-999.0_sp
              allocate(gamma_est_sim(nsim))
              gamma_est_sim=-999.0_sp
              do isim=1,nsim
                 print *
                 print *, ' -----------------------------------------------------------'
                 print *, ' Monte Carlo: simulation       ', isim,' of ', nsim
                 print *, ' -----------------------------------------------------------'
testnp2:         do itestnp=1,ntestnp
                    ! at least minNpositive positive values of mean-subtracted Ysim are required for Hill estimator
                    call generateMC(t,Ysim,Nobs,tau_est_result,alpha_est_result,gamma_est_result,opt2,pickg,rng3(iseed),rng4(iseed))
                    testNpositive=count((Ysim-sum(Ysim)/Nobs) > 0.0_sp)
                    ! Test (over the full order search range) whether at least for one order value
                    ! the tail index alpha (on the simulated data) is between 0 and 2.
                    call test_alpha_initial(Ysim,Nobs,opt2,flagtMC)
                    if (flagtMC .eq. '+' .and. .not. (opt2 .eq. 1 .and. testNpositive .lt. minNpositive)) exit testnp2
                 end do testnp2
                 call est_out(estflag,outflag,mctauflag,tau_est_result,t,Ysim,torig,Yorig,Nobs,Norig,opt2,      &
                              inputfile,dummy_MC_k_opt       ,dummy_MC_K_work  ,nsim,                           &
                              tau_est_sim(isim),tau_est_RMSE_result,alpha_est_sim(isim),gamma_est_sim(isim),    &
                              alpha_est_RMSE_result,gamma_est_RMSE_result)
              end do
              if (mctauflag .eq. 'y')                                                                           &
                 call calc_RMSE(  tau_est_sim,nsim,  tau_est_result,  tau_est_RMSE_result)
              if (opt2 .eq. 2 .and. pickg .eq. 'y') then
                 call calc_RMSE(gamma_est_sim,nsim,gamma_est_result,gamma_est_RMSE_result)
              else
                 call calc_RMSE(alpha_est_sim,nsim,alpha_est_result,alpha_est_RMSE_result)
              end if
              estflag='n'                       ! third, output estimates and RMSE
              outflag='y'
              call    est_out(estflag,outflag,  tauflag,tau_est_result,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                              inputfile,         k_opt_result,         K_result,nsim,                           &
                              tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                              alpha_est_RMSE_result,gamma_est_RMSE_result)
              if (k_opt_result > 0) call writeykopt(t,Y,Nobs,k_opt_result) ! output analysis software ht-d-stats
              deallocate(gamma_est_sim)
              deallocate(alpha_est_sim)
              deallocate(tau_est_sim)
              deallocate(Ysim)
              if (allocated(alpha_sequence_result)) deallocate(alpha_sequence_result)
              if (allocated(RMSE_result))           deallocate(RMSE_result)
           else if (flagt .eq. '-') then
              estflag='n'                       ! no estimation possible
              outflag='y'
              call    est_out(estflag,outflag,  tauflag,dummy_tau     ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,      &
                              inputfile,         k_opt_result,         K_result,nsim,                           &
                              tau_est_result,tau_est_RMSE_result,alpha_est_result,gamma_est_result,             &
                              alpha_est_RMSE_result,gamma_est_RMSE_result)
                                                !
                                                ! note: subroutine est_out (under estflag='y') allocates (in case of not allocated)
                                                !       alpha_sequence_result, RMSE_result
                                                !
              if (allocated(alpha_sequence_result)) deallocate(alpha_sequence_result)
              if (allocated(RMSE_result))           deallocate(RMSE_result)
           end if
       end if
    else if (modus .eq. 'generate') then
       Norig=MAIN_Norig
       allocate(torig(Norig))
       call generate(torig,Norig,rng3(iseed),rng4(iseed))
    end if

    !--------------------------
    ! FUNCTIONS AND SUBROUTINES
    !--------------------------

    contains

        !---------------------------------------------------------------------------------------------
        ! Brent's search in one direction for persistence estimation (modified from Numerical Recipes)
        !---------------------------------------------------------------------------------------------

        function brent(ax,bx,cx,f,tol,xmin,xfunc,yfunc,nfunc)
            implicit none
            integer nfunc
            double precision xfunc(1:nfunc),yfunc(1:nfunc)
            integer itmax
            double precision brent,ax,bx,cx,tol,xmin,f,cgold,zeps
            external f
            parameter (itmax=100,cgold=.3819660d0,zeps=1.d-18)
            integer iter
            integer ierr
            double precision a,b,d,e,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm
            a=min(ax,cx)
            b=max(ax,cx)
            v=bx
            w=v
            x=v
            e=0.d0
            fx=f(x,xfunc,yfunc,nfunc)
            fv=fx
            fw=fx
            do iter=1,itmax
              xm=0.5d0*(a+b)
              tol1=tol*abs(x)+zeps
              tol2=2.d0*tol1
              if(abs(x-xm).le.(tol2-.5d0*(b-a))) goto 3
              if(abs(e).gt.tol1) then
                r=(x-w)*(fx-fv)
                q=(x-v)*(fx-fw)
                p=(x-v)*q-(x-w)*r
                q=2.d0*(q-r)
                if(q.gt.0.d0) p=-p
                q=abs(q)
                etemp=e
                e=d
                if(abs(p).ge.abs(.5d0*q*etemp).or.p.le.q*(a-x).or.p.ge.q*(b-x))goto 1
                d=p/q
                u=x+d
                if(u-a.lt.tol2 .or. b-u.lt.tol2) d=sign(tol1,xm-x)
                goto 2
              end if
          1   if(x.ge.xm) then
                e=a-x
              else
                e=b-x
              end if
              d=cgold*e
          2   if(abs(d).ge.tol1) then
                u=x+d
              else
                u=x+sign(tol1,d)
              end if
              fu=f(u,xfunc,yfunc,nfunc)
              if(fu.le.fx) then
                if(u.ge.x) then
                  a=x
                else
                  b=x
                end if
                v=w
                fv=fw
                w=x
                fw=fx
                x=u
                fx=fu
              else
                if(u.lt.x) then
                  a=u
                else
                  b=u
                end if
                if(fu.le.fw .or. w.eq.x) then
                  v=w
                  fv=fw
                  w=u
                  fw=fu
                else if(fu.le.fv .or. v.eq.x .or. v.eq.w) then
                  v=u
                  fv=fu
                end if
              end if
            end do
            ierr = 1
            print *, ' brent: exceed maximum iterations'
          3 xmin=x
            brent=fx
            return
        end function brent

        !------------------------------------------------------------------------
        ! Median and median of absolute distances to the median (MAD) calculation
        !------------------------------------------------------------------------

        subroutine calc_medmad(x,n,xmed,xmad)
           use nrtype
           implicit none
           integer, intent(in) :: n
           real(sp), dimension(n), intent(in) :: x
           real(sp), intent(out) :: xmed
           real(sp), intent(out) :: xmad
!
!          Calculates median and MAD.
!
           real(sp), dimension(n) :: z
           real(sp), parameter :: half=0.5_sp
           z=x
           call sort(z,z,n,1,n)
           if (mod(n,2) .eq. 0) then
              xmed=(z(n/2)+z(n/2+1))*half
           else
              xmed=z(int(n/2)+1)
           end if
           z=abs(x-xmed)
           call sort(z,z,n,1,n)
           if (mod(n,2) .eq. 0) then
              xmad=(z(n/2)+z(n/2+1))*half
           else
              xmad=z(int(n/2)+1)
           end if
        end subroutine calc_medmad

        !---------------------------
        ! Empirical RMSE calculation
        !---------------------------

        subroutine calc_RMSE(theta,n,theta_true,RMSE)
            use nrtype
            implicit none
            integer, intent(in) :: n                        ! number of MC simulations
            real(sp), dimension(n), intent(in) :: theta     ! MC result
            real(sp), intent(in) :: theta_true              ! true theta value
            real(sp), intent(out) :: RMSE                   ! empirical RMSE
!
!           Calculates empirical RMSE for the MC simulations.
!            
            real(sp), parameter :: two=2.0_sp
            RMSE=sqrt(sum((theta-theta_true)**two)/n)
        end subroutine calc_RMSE

        !----------------------------------
        ! Empirical robust RMSE calculation
        !----------------------------------

        subroutine calc_RMSErob(theta,n,theta_true,RMSErob)
            use nrtype
            implicit none
            integer, intent(in) :: n                        ! number of MC simulations
            real(sp), dimension(n), intent(in) :: theta     ! MC result
            real(sp), intent(in) :: theta_true              ! true theta value
            real(sp), intent(out) :: RMSErob                ! empirical robust RMSE
!
!           Calculates empirical RMSErob for the MC simulations.
!
!           RMSErob is a robust version of the RMSE that is based on:
!           (1) the median (instead of the mean) for bias calculation and
!           (2) the normalized MAD (instead of the standard deviation) for standard error calculation.
!
!           Note that MAD is normalized: a standard Gaussian distribution has an MAD of ~0.6745.
!           The normalized MAD may hence be denoted as "robust standard deviation".
!
!           See Mudelsee (2014 Climate Time Series Analysis, 2nd edn., Springer, Cham: Table 1.2 and Eq. 3.3 therein).
!
            real(sp) :: thetamed=-999.0_sp                  ! median
            real(sp) :: thetamad=-999.0_sp                  ! normalized MAD
            real(sp), parameter :: two=2.0_sp
            real(sp), parameter :: madnorm=0.6745_sp        ! normalizing constant for MAD
            call calc_medmad(theta,n,thetamed,thetamad)
            thetamad=thetamad/madnorm
            RMSErob=sqrt((thetamed-theta_true)**two+thetamad**two)
        end subroutine calc_RMSErob

        !--------------------
        ! Cauchy distribution
        !--------------------

        subroutine cauchy(x,rng3)
        use nrtype
        use rng
        implicit none
        real(sp), intent(out) :: x
        type(rng_t), dimension(1), intent(inout) :: rng3 ! seeded random number generator
!
!       Cauchy standard distribution using the algorithm after Ahrens and Dieter (1988).
!       Ahrens JH, Dieter U (1988) Comm. ACM 31:1330.
!
        real(sp) :: S                           ! auxiliary variable
        real(sp) :: T                           ! auxiliary variable
        real(sp) :: U                           ! auxiliary variable
        real(sp) :: Uprime                      ! auxiliary variable
        real(sp), parameter ::  zero    = 0.0_sp
        real(sp), parameter ::  quarter = 0.25_sp
        real(sp), parameter ::  half    = 0.5_sp
        real(sp), parameter ::  one     = 1.0_sp
        real(sp), parameter ::  two     = 2.0_sp
        real(sp), parameter ::  a       = 0.6380631366077803_sp
        real(sp), parameter ::  b       = 0.5959486060529070_sp
        real(sp), parameter ::  q       = 0.9339962957603656_sp
        real(sp), parameter ::  W       = 0.2488702280083841_sp
        real(sp), parameter ::  aCAP    = 0.6366197723675813_sp
        real(sp), parameter ::  bCAP    = 0.5972997593539963_sp
        real(sp), parameter ::  H       = 0.0214949004570452_sp
        real(sp), parameter ::  P       = 4.9125013953033204_sp
        integer :: iseed=1                      ! dummy integer for random number generator
        U=mzran(rng3(iseed))
        T=U-half
        S=W-T**two
        if (S .gt. zero) then
           x=T*(aCAP/S+bCAP)
           return
        end if
loop:   do
          U=mzran(rng3(iseed))
          Uprime=mzran(rng3(iseed))
          T=U-half
          S=quarter-T**two
          x=T*(a/S+b)
          if (S**two*((one+x**two)*(H*Uprime+P)-q)+S .le. half) exit loop
        end do loop
        end subroutine cauchy

        !----------------------
        ! MAIN MENU
        !----------------------

        integer function choices_main(dataloaded)
            use datafilename
            implicit none
            logical,intent(inout) :: dataloaded
            integer :: opt
            do
                print '(a,a)', ' Current data file: ',trim(adjustl(inputfile))
                print *
                print '(a)',   ' Main menu                                                    [ENTER A NUMBER]'
                print *
                print '(a)',   ' 1. Data file'
                print '(a)',   ' 2. Preprocessing'
                print '(a)',   ' 3. Tail index estimation'
                print '(a)',   ' 4. Generation of artificial time series'
                print '(a)',   ' 5. Exit'
                read *, opt
                call clearscreen
                if (opt==1) dataloaded=.true.
                if ((opt>=1 .and. opt<=5) .and. dataloaded==.true.) exit
                if (opt<=5) then
                    print '(a)', ' PLEASE LOAD DATA FIRST'
                else
                    print '(a)', ' PLEASE TYPE IN A NUMBER BETWEEN 1 AND 5'
                end if
            end do
            choices_main=opt
        end function choices_main

        !--------------
        ! Clears screen
        !--------------

        subroutine clearscreen
        implicit none
        integer :: i
        integer, parameter :: n=120
        do i=1,n
           print *
        end do
        end subroutine clearscreen

        !-------------------------------------
        ! SUBROUTINE FOR ESTIMATION AND OUTPUT
        !-------------------------------------

        subroutine est_out(estflag,outflag,tauflag,tau_MC        ,t,Y   ,torig,Yorig,Nobs,Norig,opt2,                       &
                           inputfile,         k_opt,                K_work,  nsim,                                          &
                           tau_est_work  ,tau_est_RMSE_work  ,alpha_est_work  ,gamma_est_work  ,                            &
                           alpha_est_RMSE_work,  gamma_est_RMSE_work)
            use nrtype
            use cfgseedparameters, only: modus,SUBR_EST_OUT_outputfile1,SUBR_EST_OUT_outputfile2
            use Hill_Pick_testNpos_alpha
            use qbfparameters
            use rng
            use result1, only: alpha_sequence_result,RMSE_result
            use version
            implicit none
            character(len=1), intent(in)           :: estflag   ! estflag = 'y': estimation
                                                                ! estflag = 'n': no estimation
            character(len=1), intent(in)           :: outflag   ! outflag = 'y': output
                                                                ! outflag = 'n': no output
            character(len=1), intent(in)           :: tauflag   ! tauflag = 'y': persistence time estimation
                                                                ! tauflag = 'n': no persistence time estimation
            real(sp), intent(in)                   :: tau_MC    ! MC (tauflag = 'n'): do not re-estimate persistence time but overtake result
            integer, intent(in)                    :: Nobs
            integer, intent(in)                    :: Norig
            real(sp), dimension(Nobs), intent(in)  :: t         ! time (possibly preprocessed)
            real(sp), dimension(Nobs), intent(in)  :: Y         ! variable (possibly preprocessed)
            real(sp), dimension(Norig), intent(in) :: torig     ! time (original, for output)
            real(sp), dimension(Norig), intent(in) :: Yorig     ! variable (original, for output)
            integer, intent(in)                    :: opt2      ! opt2 = 1: Hill
                                                                ! opt2 = 2: Pickands
            character*1000, intent(in) :: inputfile             ! input data file name (for output)
            integer, intent(inout)  :: k_opt                    ! result: optimal order
            integer, intent(inout)  :: K_work                   ! result: K_work - 1 = maximum possible search value for the order
            integer, intent(in)     :: nsim                     ! number of MC simulations for RMSE calculations (for output)
            real(sp), intent(inout) :: tau_est_work             ! result: estimated persistence time
            real(sp), intent(in)    :: tau_est_RMSE_work        ! result: RMSE(tau_est_work), determined outside using MC
            real(sp), intent(inout) :: alpha_est_work           ! result: estimated tail index (alpha)
            real(sp), intent(inout) :: gamma_est_work           ! result: estimated tail index (gamma)
            real(sp), intent(in)    :: alpha_est_RMSE_work      ! result: RMSE(alpha_est_work), determined outside using MC
            real(sp), intent(in)    :: gamma_est_RMSE_work      ! result: RMSE(gamma_est_work), determined outside using MC
!
!           Persistence time estimation using least squares;
!
!           tail index estimation using Hill or Pickands estimator combined with an
!           order selector using brute force or quasi-brute force;
!
!           output.
!
!           Uses flags to accomodate the different tasks.
!
            character(len=80) :: outfile                                ! output file, full content
            character(len=80) :: outfile_red                            ! output file, one line of reduced content (modus 'estimate')
            character(len=9), parameter :: outfilealpha='alpha.dat'     ! output file (output analysis software ht-d-stats), alpha
            character(len=8), parameter :: outfilekopt='kopt.dat'       ! output file (output analysis software ht-d-stats), kopt
            character(len=8), parameter :: outfiledata='data.dat'       ! output file (output analysis software ht-d-stats), data (t, Y)
            character(len=1000) :: dummyc1                              ! dummy
            real(sp), dimension(:), allocatable :: Y_inner, w, Y_est    ! generated process based on the original Y
            real(sp), dimension(:), allocatable :: alpha_sequence       ! sequence of tail index estimates for all possible k values
            real(sp), dimension(:), allocatable :: RMSE                 ! RMSE measure for order selection
            real(sp), allocatable, dimension(:) :: Yordered             ! auxiliary array (used for output only here in this subroutine)
            real(sp) :: alpha_rmse                                      ! optimal tail index estimate using brute force
            real(sp) :: alpha_output_low                                ! output: lower bound of tail index estimate (estimate minus one * RMSE)
            real(sp) :: alpha_output_upp                                ! output: upper bound of tail index estimate (estimate plus  one * RMSE)
!
            real(sp) :: MaxminRMSE                                      ! variables used to store and test which k has minimum RMSE measure for order selection
            real(sp), dimension(N_inner):: RMSE_store
!
            real(sp) :: alpha_est                                       ! tail index estimate for a given k value
            real(sp) :: alpha_sim                                       ! tail index value used for order selection
!
            real(sp) :: tau_est, phi                                    ! estimate of the persistence in AR(1) model
!
            type(rng_t), dimension(N_inner), save :: rng3, rng4         ! seeds of random number generator
!
            integer :: K                                                ! K - 1 = maximum possible search value for the order
!
            integer :: Npositive                                        ! number of positive values
            integer :: N4                                               ! Pickands estimator, size of alpha_sequence; K = N4 + 1
!
            integer :: i, i_K, i_inner, count_K, counter_QBF1, J_K      ! counters
            integer :: iocheck=-999
            integer :: N_K                                              ! quasi-brute force method, number of observations for Step 1
            integer :: P_K                                              ! quasi-brute force method, number of observations for Step 2
!
            real(sp) :: rhoout, junk                                    ! auxiliary variables
            integer :: PosMaxmin                                        ! auxiliary variable
!
            real(sp), dimension(:), allocatable :: auxRMSE              ! quasi-brute force method, for storage of intermediate results
            integer, dimension(:), allocatable :: positions             ! quasi-brute force method, for storage of intermediate results
!
            real(sp), parameter :: zero=0.0_sp
            real(sp), parameter :: one=1.0_sp
!
! 1.        Estimation flag (start)
!           =======================
!
            outfile=trim(adjustl(SUBR_EST_OUT_outputfile1))             ! has to appear earlier in case estflag = 'n' and test_alpha_initial failed
            if (estflag .eq. 'y') then
!
! 2.        Preliminary
!           ===========
!
            allocate(Y_est(Nobs))
            Npositive=-999
            N4=-999
            alpha_est=-999.0_sp
            alpha_sim=-999.0_sp
            alpha_output_low=-999.0_sp
            alpha_output_upp=-999.0_sp
!
! 3.        Subtract mean and estimate persistence
!           ======================================
!
            Y_est=Y-sum(Y)/Nobs
            if (tauflag .eq. 'y') then
               call tauest(t,Y_est,Nobs,tau_est,rhoout)
            else
               tau_est=tau_MC
            end if
!
! 4.        Estimate the alpha sequence for each K value, make initializations, allocate space
!           ==================================================================================
!
            if (opt2 .eq. 1) then       ! HILL ESTIMATOR
!                                       ! alpha_sequence(i) and RMSE(i) are defined for i from 1 up to
!                                       ! the number of positive values minus 1
!                                       ! (Resnick (2007 Heavy-Tail Phenomena: Probabilistic and
!                                       ! Statistical Modeling, Springer, New York: pp. 364 therein) 
                K=count(Y_est>0.0_sp)
!
                if (K .lt. minNpositive) then
                    print '(a,i1,a)', ' Subroutine est_out: the time series has too few (',K,') positive values for executing the Hill estimator.'
                    print *
                    print '(a)',      ' Suggested options for re-analysis with the Hill estimator: try other preprocessing types or an increased data size.'
                    print '(a)',      ' See the source code, subroutine est_out, for explanation.'
                    print *
                    print '(a)',      ' Program terminates.'
                    stop
                end if            
!                
                allocate(alpha_sequence(K-1))
                alpha_sequence=-999.0_sp
                call hill(Y_est,Nobs,alpha_sequence,K)
                allocate(RMSE(K-1))
                RMSE=-999.0_sp
            else if (opt2 .eq. 2) then  ! PICKANDS ESTIMATOR
                N4=int(Nobs*fracp)
                allocate(alpha_sequence(N4))
                alpha_sequence=-999.0_sp
                call pickands(Y_est,Nobs,alpha_sequence,N4)
                allocate(RMSE(N4))
                RMSE=-999.0_sp
                K=N4+1
            end if
            k_opt=-999
            MaxminRMSE=999999999999.0_sp
            RMSE_store=-999.0_sp
            if (allocated(Y_est)) deallocate(Y_est)
            allocate(Y_inner(Nobs))
            allocate(w(Nobs))
!
! 5.        Open alpha test file
!           ====================
!
!           Note: The alpha test file allows to write simulated alpha values into a file for subsequent inspection (e.g., distribution).
!                 Uncomment the following line                     (alpha test file opened),
!                           one line in subsequent section         (alpha data written),
!                           one line in another subsequent section (alpha test file closed)
!                           (hint: search for "unit=2" to locate those sections);
!                 and recompile to perform this task.
!
!           open (unit=2, file='analysis-alpha-N_inner-K-1/alpha-N_inner-K-1.dat', status='unknown', form='formatted', action='write')
!
!
! 6.        Select optimal order
!           ====================
!
            select case (optBF)
!
! 6.1       Brute force
!           ===========
!
                case (1)
                    do i_K=1,K-1
                       if (i_K .eq. 1 .or. mod(i_K,10) .eq. 0) print *, ' BRUTE FORCE: EVALUATING VALUE ', i_K,' OF ', K-1
                       alpha_est=alpha_sequence(i_K)
                       
                       
                       ! CHECK Version 2.0
                       !
                           !if (alpha_est .le. 0.0_sp .or. alpha_est .gt. 2.0_sp) cycle
                       if (alpha_est .le. 0.0_sp) cycle
                       !
                       !
                       !
                       if (opt2 .eq. 2 .and. pickg .eq. 'y') then
                          alpha_sim=1.0_sp/alpha_est
                       else
                          alpha_sim=alpha_est
                       end if
                       !$omp parallel do private(w,Y_inner,phi,i,itestnp,testNpositive,Npositive,junk,rng3,rng4)                                              
                       do i_inner=1,N_inner
                          call rng_seed(rng3(i_inner),i_inner)  ! initialization of seeds
                          call rng_seed(rng4(i_inner),N_inner-i_inner+1)
testnp3:                  do itestnp=1,ntestnp

                             ! CHECK Version 2.0
                             !
                             !call stable(alpha_sim,Y_inner,Nobs,rng3(i_inner),rng4(i_inner))
                              call expn01(alpha_sim,Y_inner,Nobs,rng3(i_inner),rng4(i_inner))

                              
                              
                              
                              
                              do i=2,Nobs
                                if (tau_est .eq. zero) then
                                    phi=zero
                                else
                                    phi=exp(-(t(i)-t(i-1))/tau_est)
                                end if
                                Y_inner(i)=phi*Y_inner(i-1)+(one-phi*phi)*Y_inner(i)  ! generating data from an unevenly spaced, heavy-tailed AR(1) noise process;
                                                                                      ! factor (one-phi*phi) ensures strict stationarity
                                                                                      ! (Mudelsee 2014 Climate Time Series Analysis, 2nd edn., Springer, Cham: Ch. 2 therein),
                                                                                      ! which may be relevant for finite-variance processes
                             end do
                             Y_inner=Y_inner-sum(Y_inner)/Nobs                        ! mean subtraction
!
!                            Note: Test for Npositive > 1. The situation Npositive = 1 may be rarely met,
!                                  but it may occur during the order selection (i.e., random simulations);
!                                  specifically, if a time series has just one very large positive value and
!                                  the other have small absolute values, then the mean subtraction may lead to
!                                  a situation where the input series to subroutine hill has just one (large)
!                                  positive value, and Nobs - 1 (small) negative values.
!                                  Then, in subroutine Hill, gamma1 may be allocated an array of extent zero,
!                                  which would lead to a run-time error when trying to assign a value to gamma1(1).
!
                             testNpositive=count(Y_inner>0.0_sp)
                             if (testNpositive .ge. minNpositive) exit testnp3
                          end do testnp3
!
! 6.1.1     Hill
!           ====
!
                          if (opt2 .eq. 1) then
                             Npositive=count(Y_inner>0.0_sp)
                             call hill(Y_inner,Nobs,w,Npositive)
                             junk=real(Npositive-1)*(real(i_K)/real(K-1))   ! because Y_inner are new random data (compared to Y),
                             i=nint(junk)                                   ! Npositive may be different from K (i.e., number of positive values of Y);
                             if (i<1) i=1                                   ! usage of junk and i corrects for that: w(i) is taken at the same
                             RMSE_store(i_inner)=w(i)                       ! relative position, real(i_K)/real(K-1)
!
! 6.1.2     Pickands
!           ========
!
                          else if (opt2 .eq. 2) then
                             call pickands(Y_inner,Nobs,w,N4)
                             junk=real(N4)*(real(i_K)/real(K-1))
                             i=nint(junk)
                             if (i<1) i=1
                             RMSE_store(i_inner)=w(i)
                          end if
                       end do
                       !$omp end parallel do
!
! 6.1.3     Write alpha data into alpha test file
!           =====================================
!
!                     write (unit=2, fmt='(1000(1x,es12.3e2))') RMSE_store(1:N_inner)  ! note that N_inner = 1000
!
! 6.1.4     Storing intermediate result
!           ===========================
!
!           Note: The search for optimal k is restricted to a meaningful range (educated guess).
!
                       RMSE(i_K)=sqrt(sum((RMSE_store-alpha_est)**2)/N_inner)
                       if (RMSE(i_K)<MaxminRMSE) then
                          k_opt=i_K
                          MaxminRMSE=RMSE(i_K)
                       end if
                    end do
!
! 6.1.5     Close alpha test file
!           =====================
!
!                  close (unit=2, status='keep')
!
! 6.1.6     Evaluating final result
!           =======================
!
                    alpha_rmse=alpha_sequence(k_opt)
!
! 6.2       Quasi-brute force
!           =================
!
!           The search consists of two steps.   Step 1:     coarse search,
!
!                                                           do i_K = 1, K-1, L_k         with L_k > 1
!
!
!
!                                               Step 2:     take from coarse search points only the
!
!                                                           PminK percent with the lowest RMSE measure for order selection,
!
!                                                           then
!
!                                                           perform a fine search around these points,
!
!                                                           do count_K = 1, P_K
!                                                              do j_K = 1, L_k
!
                case (2)
                                         !
                                         ! Step 1
                                         ! ======
                                         !
                                         ! We make intervals of lenght L_k to reduce the quantity of k values evaluated.
                                         !
                N_K=int((K-2)/L_k)+1     ! number of points of K which are used in the first step
                P_K=int(PminK*(N_K)/100) ! number of points of K which are used in the second step
                allocate(auxRMSE(P_K))   ! this array contains the lower values of RMSE
                allocate(positions(P_K)) ! this array contains the position of the lower values
                                         !
                                         ! For each L_k value of k, we repeat the estimation N_inner times.
                                         !
                auxRMSE=-999.0_sp        ! initialization
                positions=-999           ! initialization
                count_K=0
                counter_QBF1=0
                MaxminRMSE=-999999999999.0_sp
                do i_K=1,K-1,L_k
                    counter_QBF1=counter_QBF1+1
                    print *, ' QUASI-BRUTE FORCE, STEP 1/2: EVALUATING VALUE ', counter_QBF1,' OF ', N_K
                    alpha_est=alpha_sequence(i_K)
                    
                    
                    ! CHECK Version 2.0
                    !
                    !
                       !if (alpha_est .le. 0.0_sp .or. alpha_est .gt. 2.0_sp) cycle
                    if (alpha_est .le. 0.0_sp) cycle
                    !
                    !
                    !
                    count_K=count_K+1                   
                    if (opt2 .eq. 2 .and. pickg .eq. 'y') then
                       alpha_sim=1.0_sp/alpha_est
                    else
                       alpha_sim=alpha_est
                    end if
                    !$omp parallel do private(w,Y_inner,phi,i,itestnp,testNpositive,Npositive,junk,rng3,rng4)
                    do i_inner=1,N_inner
                       call rng_seed(rng3(i_inner),i_inner)  ! initialization of seeds
                       call rng_seed(rng4(i_inner),N_inner-i_inner+1)
testnp4:               do itestnp=1,ntestnp



                          ! CHECK Version 2.0
                          !
                          !
                          ! call stable(alpha_sim,Y_inner,Nobs,rng3(i_inner),rng4(i_inner))
                          call expn01(alpha_sim,Y_inner,Nobs,rng3(i_inner),rng4(i_inner))
                          !
                          !
                          !
                          do i=2,Nobs
                              if (tau_est .eq. zero) then
                                  phi=zero
                              else
                                  phi=exp(-(t(i)-t(i-1))/tau_est)
                              end if
                              Y_inner(i)=phi*Y_inner(i-1)+(one-phi*phi)*Y_inner(i)  ! generating data from an unevenly spaced, heavy-tailed AR(1) noise process;
                                                                                    ! factor (one-phi*phi) ensures strict stationarity
                                                                                    ! (Mudelsee 2014 Climate Time Series Analysis, 2nd edn., Springer, Cham: Ch. 2 therein),
                                                                                    ! which may be relevant for finite-variance processes
                          end do
                          Y_inner=Y_inner-sum(Y_inner)/Nobs                         ! mean subtraction
!
                          testNpositive=count(Y_inner>0.0_sp)
                          if (testNpositive .ge. minNpositive) exit testnp4
                       end do testnp4
!
! 6.2.1     Step 1/2, Hill
!           ==============
!
                        if (opt2 .eq. 1) then
                            Npositive=count(Y_inner>0.0_sp)
                            call hill(Y_inner,Nobs,w,Npositive)
                            junk=real(Npositive-1)*(real(i_K)/real(K-1))
                            i=nint(junk)
                            if (i<1) i=1
                            RMSE_store(i_inner)=w(i)
!
! 6.2.2     Step 1/2, Pickands
!           ==================
!
                        else if (opt2 .eq. 2) then
                            call pickands(Y_inner,Nobs,w,N4)
                            junk=real(N4)*(real(i_K)/real(K-1))
                            i=nint(junk)
                            if (i<1) i=1
                            RMSE_store(i_inner)=w(i)
                        end if
                    end do
                    !$omp end parallel do
!
! 6.2.3     Storing intermediate result
!           ===========================
!
                    RMSE(i_K)=sqrt(sum((RMSE_store-alpha_est)**2)/N_inner)
                    if (count_K .lt. P_K) then              ! First, fill up positions and auxRMSE until nominal number P_K reached
                        positions(count_K)=i_K
                        auxRMSE(count_K)=RMSE(i_K)
                    else if (count_K .eq. P_K) then         ! Second, if P_K reached, then define MaxminRMSE and PosMaxmin
                        positions(count_K)=i_K
                        auxRMSE(count_K)=RMSE(i_K)
                        MaxminRMSE=maxval(auxRMSE,dim=1)
                        PosMaxmin=maxloc(auxRMSE,dim=1)
                    end if
                    if (count_K .gt. P_K .and. RMSE(i_K) .lt. MaxminRMSE) then
                                                            !
                                                            ! Third. For the remaining trials of i_K, throw out each time the positions and auxRMSE
                                                            ! with the largest RMSE measure for order selection.
                                                            ! That means, retain the pool of size P_K of best i_K values.
                                                            !
                                                            ! After Step 1 is completed, the P_K values of i_K (with step size L_k)
                                                            ! are retained that yield smallest RMSE measure for order selection.
                                                            ! These values are taken for fine evaluation in Step 2.
                                                            !
                        positions(PosMaxmin)=i_K
                        auxRMSE(PosMaxmin)=RMSE(i_K)
                        MaxminRMSE=maxval(auxRMSE,dim=1)
                        PosMaxmin=maxloc(auxRMSE,dim=1)
                    end if
                end do
                                         !
                                         ! Step 2
                                         ! ======
                                         !
                                         ! We select the lower PminK% values of RMSE, and the experiment is repeated in their corresponding intervals.
                                         ! For each L_k value of k, we repeat the estimation N_inner times.
                                         !
                MaxminRMSE=999999999999.0_sp
                do count_K=1,P_K
                    do j_K=1,L_k
                        print *, ' QUASI-BRUTE FORCE, STEP 2/2: EVALUATING VALUE ', (count_K-1)*L_k+j_K,' OF ', P_K*L_k
                        i_K=positions(count_K)+j_K-1
                        if (i_K>size(alpha_sequence)) i_K=size(alpha_sequence)
                        alpha_est=alpha_sequence(i_K)
                        
                        
                        
                        ! CHECK Version 2.0
                        !
                        !
                           !if (alpha_est .le. 0.0_sp .or. alpha_est .gt. 2.0_sp) cycle
                        if (alpha_est .le. 0.0_sp) cycle
                        !
                        !
                        !
                        if (opt2 .eq. 2 .and. pickg .eq. 'y') then
                           alpha_sim=1.0_sp/alpha_est
                        else
                           alpha_sim=alpha_est
                        end if
                        !$omp parallel do private(w,Y_inner,phi,i,itestnp,testNpositive,Npositive,junk,rng3,rng4)
                        do i_inner=1,N_inner
                           call rng_seed(rng3(i_inner),i_inner)  ! initialization of seeds
                           call rng_seed(rng4(i_inner),N_inner-i_inner+1)
testnp5:                   do itestnp=1,ntestnp



                              ! CHECK Version 2.0
                              !
                              !
                              !     call stable(alpha_sim,Y_inner,Nobs,rng3(i_inner),rng4(i_inner))
                              call expn01(alpha_sim,Y_inner,Nobs,rng3(i_inner),rng4(i_inner))
                              !
                              !
                              !
                              do i=2,Nobs
                                 if (tau_est .eq. zero) then
                                     phi=zero
                                 else
                                     phi=exp(-(t(i)-t(i-1))/tau_est)
                                 end if
                                 Y_inner(i)=phi*Y_inner(i-1)+(one-phi*phi)*Y_inner(i)  ! generating data from an unevenly spaced, heavy-tailed AR(1) noise process;
                                                                                       ! factor (one-phi*phi) ensures strict stationarity
                                                                                       ! (Mudelsee 2014 Climate Time Series Analysis, 2nd edn., Springer, Cham: Ch. 2 therein),
                                                                                       ! which may be relevant for finite-variance processes
                              end do
                              Y_inner=Y_inner-sum(Y_inner)/Nobs                         ! mean subtraction
!
                              testNpositive=count(Y_inner>0.0_sp)
                              if (testNpositive .ge. minNpositive) exit testnp5
                           end do testnp5
!
! 6.2.4     Step 2/2, Hill
!           ==============
!
                            if (opt2 .eq. 1) then
                                Npositive=count(Y_inner>0.0_sp)
                                call hill(Y_inner,Nobs,w,Npositive)
                                junk=real(Npositive-1)*(real(i_K)/real(K-1))
                                i=nint(junk)
                                if (i<1) i=1
                                RMSE_store(i_inner)=w(i)
!
! 6.2.5     Step 2/2, Pickands
!           ==================
!
                            else if (opt2 .eq. 2) then
                               call pickands(Y_inner,Nobs,w,N4)
                               junk=real(N4)*(real(i_K)/real(K-1))
                               i=nint(junk)
                               if (i<1) i=1
                               RMSE_store(i_inner)=w(i)
                            end if
                        end do
                        !$omp end parallel do
!
! 6.2.6     Storing intermediate result
!           ===========================
!
!           Note: The search for optimal k is restricted to a meaningful range (educated guess).
!
                        RMSE(i_K)=sqrt(sum((RMSE_store-alpha_est)**2)/N_inner)
                        if (RMSE(i_K)<MaxminRMSE) then
                            k_opt=i_K
                            MaxminRMSE=RMSE(i_K)
                        end if
                    end do
                end do
!
! 6.2.7     Evaluating final result
!           =======================
!
                alpha_rmse=alpha_sequence(k_opt)
!
! 6.2.8     Deallocate arrays used for case (2)
!           ===================================
!
                if (allocated(auxRMSE)) deallocate(auxRMSE)
                if (allocated(positions)) deallocate(positions)
            end select
!
! 6.2.9     Deallocate other arrays
!           =======================
!
            if (allocated(Y_inner)) deallocate(Y_inner)
            if (allocated(w)) deallocate(w)
!
! 7.        Result
!           ======
!
            tau_est_work=tau_est
            K_work=K
            alpha_est_work=-999.0_sp
            gamma_est_work=-999.0_sp
            if (opt2 .eq. 2 .and. pickg .eq. 'y') then
               gamma_est_work=alpha_rmse
            else
               alpha_est_work=alpha_rmse
            end if
            if (.not. allocated(alpha_sequence_result)) then    ! first called, subroutine est_out estimates using the time series
               allocate(alpha_sequence_result(K-1))             ! (not the MC series) and stores result
               alpha_sequence_result=alpha_sequence             ! later (during MC and at output), is is not being overwritten
            end if
            if (.not. allocated(RMSE_result)) then              ! first called, subroutine est_out estimates using the time series
               allocate(RMSE_result(K-1))                       ! (not the MC series) and stores result
               RMSE_result=RMSE                                 ! later (during MC and at output), is is not being overwritten
            end if
!
! 8.        Estimation flag (end)
!           =====================
!
            end if
!
! 9.        Output
!           ======
!
! 9.1       Output flag (start)
!           ===================
!
            if (outflag .eq. 'y') then
!
! 9.2       Screen
!           ======
!
            call clearscreen
            print '(a)',             ' Optimal Heavy Tail Estimation: Results'
            print '(a)',             ' ============================='
            print '(a)',             ' '
            print '(a)',             ' Heavy tail index estimation (Hill or Pickands estimator)'
            print '(a)',             ' on time series data by means of a brute force order selector'
            print '(a)',             ' (RMSE measure), including preprocessing routines.'
            print '(a)',             ' '
            print '(a,a)',           ' Version ',versiondate
            print '(a)',             ' '
            print '(a)',             ' Copyright (C) Climate Risk Analysis - Manfred Mudelsee'
            print '(a)',             ' http://www.climate-risk-analysis.com'
            print '(a)',             ' '
            print '(a)',             ' Please see source code (ht.f90) for disclaimer and permissions.'
            print '(a)',             ' '
            print '(a,a)',           ' Input data file name:                                                    ',trim(adjustl(inputfile))
            if (Norig .lt. 100) then
               print '(a,i2)',       ' Data size (original):                                                    ',Norig
            else if (Norig .ge. 100 .and. Norig .lt. 1000) then
               print '(a,i3)',       ' Data size (original):                                                    ',Norig
            else if (Norig .ge. 1000 .and. Norig .lt. 10000) then
               print '(a,i4)',       ' Data size (original):                                                    ',Norig
            else if (Norig .ge. 10000 .and. Norig .lt. 100000) then
               print '(a,i5)',       ' Data size (original):                                                    ',Norig
            else if (Norig .ge. 100000) then
               print '(a,i6)',       ' Data size (original):                                                    ',Norig
            end if
            if (Nobs .lt. 100) then
               print '(a,i2)',       ' Data size (after preprocessing):                                         ',Nobs
            else if (Nobs .ge. 100 .and. Nobs .lt. 1000) then
               print '(a,i3)',       ' Data size (after preprocessing):                                         ',Nobs
            else if (Nobs .ge. 1000 .and. Nobs .lt. 10000) then
               print '(a,i4)',       ' Data size (after preprocessing):                                         ',Nobs
            else if (Nobs .ge. 10000 .and. Nobs .lt. 100000) then
               print '(a,i5)',       ' Data size (after preprocessing):                                         ',Nobs
            else if (Nobs .ge. 100000) then
               print '(a,i6)',       ' Data size (after preprocessing):                                         ',Nobs
            end if
            if (opt2 .eq. 1) then
               print '(a,a)',        ' Tail index (alpha) estimator:                                            ','Hill'
            else if (opt2 .eq. 2) then
               if (pickg .eq. 'y') then
                  print '(a,a)',     ' Tail index (gamma) estimator:                                            ','Pickands'
               else if (pickg .eq. 'n') then
                  print '(a,a)',     ' Tail index (alpha) estimator:                                            ','Pickands'
               end if
            end if
            if (optBF .eq. 1) then
               print '(a,a)',        ' Order selector (RMSE measure):                                           ','Brute force'
               print '(a,a)',        ' Length of search-steps used in Step 1:                             L_k = ','Not applicable (only for quasi-brute force)'
               print '(a,a)',        ' Percentage of best k values used in Step 2:                      Pmink = ','Not applicable (only for quasi-brute force)'
            else if (optBF .eq. 2) then
               print '(a,a)',        ' Order selector (RMSE measure):                                           ','Quasi-brute force (in 2 Steps)'
               if (L_k .lt. 10) then
                  print '(a,i1)',    ' Length of search-steps used in Step 1:                             L_k = ',L_k
               else if (L_k .ge. 10 .and. L_k .lt. 100) then
                  print '(a,i2)',    ' Length of search-steps used in Step 1:                             L_k = ',L_k
               else if (L_k .ge. 100 .and. L_k .lt. 1000) then
                  print '(a,i3)',    ' Length of search-steps used in Step 1:                             L_k = ',L_k
               else if (L_k .ge. 1000 .and. L_k .lt. 10000) then
                  print '(a,i4)',    ' Length of search-steps used in Step 1:                             L_k = ',L_k
               else if (L_k .ge. 10000) then
                  print '(a,i5)',    ' Length of search-steps used in Step 1:                             L_k = ',L_k
               end if
               if (Pmink .lt. 10) then
                  print '(a,i1)',    ' Percentage of best k values used in Step 2:                      Pmink = ',Pmink
               else if (Pmink .ge. 10 .and. Pmink .lt. 100) then
                  print '(a,i2)',    ' Percentage of best k values used in Step 2:                      Pmink = ',Pmink
               else if (Pmink .eq. 100) then
                  print '(a,i3)',    ' Percentage of best k values used in Step 2:                      Pmink = ',Pmink
               end if               
            end if
            if (N_inner .lt. 1000) then
               print '(a,i3)',       ' Number of loops (order selection):                                       ',N_inner
            else if (N_inner .ge. 1000 .and. N_inner .lt. 10000) then
               print '(a,i4)',       ' Number of loops (order selection):                                       ',N_inner
            else if (N_inner .ge. 10000 .and. N_inner .lt. 100000) then
               print '(a,i5)',       ' Number of loops (order selection):                                       ',N_inner
            else if (N_inner .ge. 100000) then
               print '(a,i6)',       ' Number of loops (order selection):                                       ',N_inner
            end if
            if (nsim .lt. 0) then
               print '(a,a)',        ' Number of Monte Carlo simulations for RMSE calculations:                 ','(No simulations done)'
            else if (nsim .ge. 10 .and. nsim .lt. 100) then
               print '(a,i2)',       ' Number of Monte Carlo simulations for RMSE calculations:                 ',nsim
            else if (nsim .ge. 100 .and. nsim .lt. 1000) then
               print '(a,i3)',       ' Number of Monte Carlo simulations for RMSE calculations:                 ',nsim
            else if (nsim .ge. 1000 .and. nsim .lt. 10000) then
               print '(a,i4)',       ' Number of Monte Carlo simulations for RMSE calculations:                 ',nsim
            else if (nsim .ge. 10000 .and. nsim .lt. 100000) then
               print '(a,i5)',       ' Number of Monte Carlo simulations for RMSE calculations:                 ',nsim
            else if (nsim .ge. 100000) then
               print '(a,i6)',       ' Number of Monte Carlo simulations for RMSE calculations:                 ',nsim
            end if
            print '(a)',       ' '
            if (tau_est_work .ge. 0.0_sp .and. tau_est_work .lt. 10.0_sp) then
               print '(a,f8.6)',     ' Persistence time estimate:                                        tau  = ',tau_est_work
            else if (tau_est_work .ge. 10.0_sp .and. tau_est_work .lt. 100.0_sp) then
               print '(a,f9.6)',     ' Persistence time estimate:                                        tau  = ',tau_est_work
            else if (tau_est_work .ge. 100.0_sp .and. tau_est_work .lt. 1000.0_sp) then
               print '(a,f10.6)',    ' Persistence time estimate:                                        tau  = ',tau_est_work
            else
               print '(a,f11.6)',    ' Persistence time estimate:                                        tau  = ',tau_est_work
            end if
            if (tau_est_RMSE_work .ge. 0.0_sp .and. tau_est_RMSE_work .lt. 10.0_sp) then
               print '(a,f8.6)',     ' Error:                                                       RMSE(tau) = ',tau_est_RMSE_work
             else if (tau_est_RMSE_work .ge. 10.0_sp .and. tau_est_RMSE_work .lt. 100.0_sp) then
               print '(a,f9.6)',     ' Error:                                                       RMSE(tau) = ',tau_est_RMSE_work
            else if (tau_est_RMSE_work .ge. 100.0_sp .and. tau_est_RMSE_work .lt. 1000.0_sp) then
               print '(a,f10.6)',    ' Error:                                                       RMSE(tau) = ',tau_est_RMSE_work
            else
               print '(a,f11.6)',    ' Error:                                                       RMSE(tau) = ',tau_est_RMSE_work
            end if
            if (opt2 .eq. 2 .and. pickg .eq. 'y') then
               if (gamma_est_work .ge. 0.0_sp .and. gamma_est_work .lt. 10.0_sp) then
                  print '(a,f8.6)',  ' Tail index estimate:                                            gamma  = ',gamma_est_work
               else if (gamma_est_work .ge. 10.0_sp .and. gamma_est_work .lt. 100.0_sp) then
                  print '(a,f9.6)',  ' Tail index estimate:                                            gamma  = ',gamma_est_work
               else if (gamma_est_work .ge. 100.0_sp .and. gamma_est_work .lt. 1000.0_sp) then
                  print '(a,f10.6)', ' Tail index estimate:                                            gamma  = ',gamma_est_work
               else
                  print '(a,f11.6)', ' Tail index estimate:                                            gamma  = ',gamma_est_work
               end if
               if (gamma_est_RMSE_work .ge. 0.0_sp .and. gamma_est_RMSE_work .lt. 10.0_sp) then
                  print '(a,f8.6)',  ' Error:                                                     RMSE(gamma) = ',gamma_est_RMSE_work
               else if (gamma_est_RMSE_work .ge. 10.0_sp .and. gamma_est_RMSE_work .lt. 100.0_sp) then
                  print '(a,f9.6)',  ' Error:                                                     RMSE(gamma) = ',gamma_est_RMSE_work
               else if (gamma_est_RMSE_work .ge. 100.0_sp .and. gamma_est_RMSE_work .lt. 1000.0_sp) then
                  print '(a,f10.6)', ' Error:                                                     RMSE(gamma) = ',gamma_est_RMSE_work
               else
                  print '(a,f11.6)', ' Error:                                                     RMSE(gamma) = ',gamma_est_RMSE_work
               end if
            else
               if (alpha_est_work .ge. 0.0_sp .and. alpha_est_work .lt. 10.0_sp) then
                  print '(a,f8.6)',  ' Tail index estimate:                                            alpha  = ',alpha_est_work
               else if (alpha_est_work .ge. 10.0_sp .and. alpha_est_work .lt. 100.0_sp) then
                  print '(a,f9.6)',  ' Tail index estimate:                                            alpha  = ',alpha_est_work
               else if (alpha_est_work .ge. 100.0_sp .and. alpha_est_work .lt. 1000.0_sp) then
                  print '(a,f10.6)', ' Tail index estimate:                                            alpha  = ',alpha_est_work
               else
                  print '(a,f11.6)', ' Tail index estimate:                                            alpha  = ',alpha_est_work
               end if
               if (alpha_est_RMSE_work .ge. 0.0_sp .and. alpha_est_RMSE_work .lt. 10.0_sp) then
                  print '(a,f8.6)',  ' Error:                                                     RMSE(alpha) = ',alpha_est_RMSE_work
               else if (alpha_est_RMSE_work .ge. 10.0_sp .and. alpha_est_RMSE_work .lt. 100.0_sp) then
                  print '(a,f9.6)',  ' Error:                                                     RMSE(alpha) = ',alpha_est_RMSE_work
               else if (alpha_est_RMSE_work .ge. 100.0_sp .and. alpha_est_RMSE_work .lt. 1000.0_sp) then
                  print '(a,f10.6)', ' Error:                                                     RMSE(alpha) = ',alpha_est_RMSE_work
               else
                  print '(a,f11.6)', ' Error:                                                     RMSE(alpha) = ',alpha_est_RMSE_work
               end if
            end if

            if (k_opt .lt. 1) then
               print '(a,i4)',       ' Optimal order:                                                   k_opt = ',k_opt
            else if (k_opt .ge. 1     .and. k_opt .lt. 10) then
               print '(a,i1)',       ' Optimal order:                                                   k_opt = ',k_opt
            else if (k_opt .ge. 10    .and. k_opt .lt. 100) then
               print '(a,i2)',       ' Optimal order:                                                   k_opt = ',k_opt
            else if (k_opt .ge. 100   .and. k_opt .lt. 1000) then
               print '(a,i3)',       ' Optimal order:                                                   k_opt = ',k_opt
            else if (k_opt .ge. 1000  .and. k_opt .lt. 10000) then
               print '(a,i4)',       ' Optimal order:                                                   k_opt = ',k_opt
            else if (k_opt .ge. 10000 .and. k_opt .lt. 100000) then
               print '(a,i5)',       ' Optimal order:                                                   k_opt = ',k_opt
            else if (k_opt .ge. 100000) then
               print '(a,i6)',       ' Optimal order:                                                   k_opt = ',k_opt
            end if
            if (alpha_est_work .eq. -999.0_sp) then
               print *
               print '(a,a)',        ' Note: alpha = -999.0 means that no estimation is possible since for all k,'
               print '(a,a)',        '       alpha(k) is outside of the interval [0; 2] (exclusive of the endpoints).'
               print '(a,a)',        '       Suggested options for re-analysis:'
               print '(a,a)',        '       try other preprocessing types or an increased data size.'
            end if
!
! 9.3       Output file, full content
!           =========================
!
!           Notes: (1) modus 'per-hand'
!                      full content is written in file = SUBR_EST_OUT_outputfile1,
!                      set in module cfgseedparameters;
!                  (2) modus 'estimate'
!                      full content is written in file = SUBR_EST_OUT_outputfile1,
!                      set in the configuration file;
!                      via a batch file consisting of many lines 'ht.exe ht001.cfg'
!                                                                'ht.exe ht002.cfg'
!                                                                'ht.exe ht003.cfg',
!                                                                etc.,
!                      where each configuration file sets own    SUBR_READ_FILE_inputfile,
!                                                                SUBR_EST_OUT_outputfile1,
!                      it is possible to estimate alpha
!                      automatically for many input data
!                      and write full output content in
!                      different files.
!
!                      If an 'external' MC experiment is performed,
!                      then SUBR_EST_OUT_outputfile2
!                      with one line of reduced content per
!                      analysed series should be of interest
!                      (see Section 9.4).
!
            open (unit=1, file=outfile, status='unknown', form='formatted', action='write')
            write (unit=1, fmt='(a)')             'Optimal Heavy Tail Estimation: Results'
            write (unit=1, fmt='(a)')             '============================='
            write (unit=1, fmt='(a)')             ' '
            write (unit=1, fmt='(a)')             'Heavy tail index estimation (Hill or Pickands estimator)'
            write (unit=1, fmt='(a)')             'on time series data by means of a brute force order selector'
            write (unit=1, fmt='(a)')             '(RMSE measure), including preprocessing routines.'
            write (unit=1, fmt='(a)')             ' '
            write (unit=1, fmt='(a,a)')           ' Version ',versiondate
            write (unit=1, fmt='(a)')             ' '
            write (unit=1, fmt='(a)')             'Copyright (C) Climate Risk Analysis - Manfred Mudelsee'
            write (unit=1, fmt='(a)')             'http://www.climate-risk-analysis.com'
            write (unit=1, fmt='(a)')             ' '
            write (unit=1, fmt='(a)')             'Please see source code (ht.f90) for disclaimer and permissions.'
            write (unit=1, fmt='(a)')             ' '
            write (unit=1, fmt='(a,a)')           'Input data file name:                                                   ',trim(adjustl(inputfile))
            if (Norig .lt. 100) then
               write (unit=1, fmt='(a,i2)')       'Data size (original):                                                   ',Norig
            else if (Norig .ge. 100 .and. Norig .lt. 1000) then
               write (unit=1, fmt='(a,i3)')       'Data size (original):                                                   ',Norig
            else if (Norig .ge. 1000 .and. Norig .lt. 10000) then
               write (unit=1, fmt='(a,i4)')       'Data size (original):                                                   ',Norig
            else if (Norig .ge. 10000 .and. Norig .lt. 100000) then
               write (unit=1, fmt='(a,i5)')       'Data size (original):                                                   ',Norig
            else if (Norig .ge. 100000) then
               write (unit=1, fmt='(a,i6)')       'Data size (original):                                                   ',Norig
            end if
            if (Nobs .lt. 100) then
               write (unit=1, fmt='(a,i2)')       'Data size (after preprocessing):                                        ',Nobs
            else if (Nobs .ge. 100 .and. Nobs .lt. 1000) then
               write (unit=1, fmt='(a,i3)')       'Data size (after preprocessing):                                        ',Nobs
            else if (Nobs .ge. 1000 .and. Nobs .lt. 10000) then
               write (unit=1, fmt='(a,i4)')       'Data size (after preprocessing):                                        ',Nobs
            else if (Nobs .ge. 10000 .and. Nobs .lt. 100000) then
               write (unit=1, fmt='(a,i5)')       'Data size (after preprocessing):                                        ',Nobs
            else if (Nobs .ge. 100000) then
               write (unit=1, fmt='(a,i6)')       'Data size (after preprocessing):                                        ',Nobs
            end if
            if (opt2 .eq. 1) then
               write (unit=1, fmt='(a,a)')        'Tail index (alpha) estimator:                                           ','Hill'
            else if (opt2 .eq. 2) then
               if (pickg .eq. 'y') then
                  write (unit=1, fmt='(a,a)')     'Tail index (gamma) estimator:                                           ','Pickands'
               else if (pickg .eq. 'n') then
                  write (unit=1, fmt='(a,a)')     'Tail index (alpha) estimator:                                           ','Pickands'
               end if
            end if
            if (optBF .eq. 1) then
               write (unit=1, fmt='(a,a)')        'Order selector (RMSE measure):                                          ','Brute force'
               write (unit=1, fmt='(a,a)')        'Length of search-steps used in Step 1:                            L_k = ','Not applicable (only for quasi-brute force)'
               write (unit=1, fmt='(a,a)')        'Percentage of best k values used in Step 2:                     Pmink = ','Not applicable (only for quasi-brute force)'
            else if (optBF .eq. 2) then
               write (unit=1, fmt='(a,a)')        'Order selector (RMSE measure):                                          ','Quasi-brute force (in 2 Steps)'
               if (L_k .lt. 10) then
                  write (unit=1, fmt='(a,i1)')    'Length of search-steps used in Step 1:                            L_k = ',L_k
               else if (L_k .ge. 10 .and. L_k .lt. 100) then
                  write (unit=1, fmt='(a,i2)')    'Length of search-steps used in Step 1:                            L_k = ',L_k
               else if (L_k .ge. 100 .and. L_k .lt. 1000) then
                  write (unit=1, fmt='(a,i3)')    'Length of search-steps used in Step 1:                            L_k = ',L_k
               else if (L_k .ge. 1000 .and. L_k .lt. 10000) then
                  write (unit=1, fmt='(a,i4)')    'Length of search-steps used in Step 1:                            L_k = ',L_k
               else if (L_k .ge. 10000) then
                  write (unit=1, fmt='(a,i5)')    'Length of search-steps used in Step 1:                            L_k = ',L_k
               end if
               if (Pmink .lt. 10) then
                  write (unit=1, fmt='(a,i1)')    'Percentage of best k values used in Step 2:                     Pmink = ',Pmink
               else if (Pmink .ge. 10 .and. Pmink .lt. 100) then
                  write (unit=1, fmt='(a,i2)')    'Percentage of best k values used in Step 2:                     Pmink = ',Pmink
               else if (Pmink .eq. 100) then
                  write (unit=1, fmt='(a,i3)')    'Percentage of best k values used in Step 2:                     Pmink = ',Pmink
               end if               
            end if
            if (N_inner .lt. 1000) then
               write (unit=1, fmt='(a,i3)')       'Number of loops (order selection):                                      ',N_inner
            else if (N_inner .ge. 1000 .and. N_inner .lt. 10000) then
               write (unit=1, fmt='(a,i4)')       'Number of loops (order selection):                                      ',N_inner
            else if (N_inner .ge. 10000 .and. N_inner .lt. 100000) then
               write (unit=1, fmt='(a,i5)')       'Number of loops (order selection):                                      ',N_inner
            else if (N_inner .ge. 100000) then
               write (unit=1, fmt='(a,i6)')       'Number of loops (order selection):                                      ',N_inner
            end if
            if (nsim .lt. 0) then
               write (unit=1, fmt='(a,a)')        'Number of Monte Carlo simulations for RMSE calculations:                ','(No MC simulations done)'
            else if (nsim .ge. 10 .and. nsim .lt. 100) then
               write (unit=1, fmt='(a,i2)')       'Number of Monte Carlo simulations for RMSE calculations:                ',nsim
            else if (nsim .ge. 100 .and. nsim .lt. 1000) then
               write (unit=1, fmt='(a,i3)')       'Number of Monte Carlo simulations for RMSE calculations:                ',nsim
            else if (nsim .ge. 1000 .and. nsim .lt. 10000) then
               write (unit=1, fmt='(a,i4)')       'Number of Monte Carlo simulations for RMSE calculations:                ',nsim
            else if (nsim .ge. 10000 .and. nsim .lt. 100000) then
               write (unit=1, fmt='(a,i5)')       'Number of Monte Carlo simulations for RMSE calculations:                ',nsim
            else if (nsim .ge. 100000) then
               write (unit=1, fmt='(a,i6)')       'Number of Monte Carlo simulations for RMSE calculations:                ',nsim
            end if
            write (unit=1, fmt='(a)')             ' '
            if (tau_est_work .ge. 0.0_sp .and. tau_est_work .lt. 10.0_sp) then
               write (unit=1, fmt='(a,f8.6)')     'Persistence time estimate:                                       tau  = ',tau_est_work
            else if (tau_est_work .ge. 10.0_sp .and. tau_est_work .lt. 100.0_sp) then
               write (unit=1, fmt='(a,f9.6)')     'Persistence time estimate:                                       tau  = ',tau_est_work
            else if (tau_est_work .ge. 100.0_sp .and. tau_est_work .lt. 1000.0_sp) then
               write (unit=1, fmt='(a,f10.6)')    'Persistence time estimate:                                       tau  = ',tau_est_work
            else
               write (unit=1, fmt='(a,f11.6)')    'Persistence time estimate:                                       tau  = ',tau_est_work
            end if
            if (tau_est_RMSE_work .ge. 0.0_sp .and. tau_est_RMSE_work .lt. 10.0_sp) then
               write (unit=1, fmt='(a,f8.6)')     'Error:                                                      RMSE(tau) = ',tau_est_RMSE_work
             else if (tau_est_RMSE_work .ge. 10.0_sp .and. tau_est_RMSE_work .lt. 100.0_sp) then
               write (unit=1, fmt='(a,f9.6)')     'Error:                                                      RMSE(tau) = ',tau_est_RMSE_work
            else if (tau_est_RMSE_work .ge. 100.0_sp .and. tau_est_RMSE_work .lt. 1000.0_sp) then
               write (unit=1, fmt='(a,f10.6)')    'Error:                                                      RMSE(tau) = ',tau_est_RMSE_work
            else
               write (unit=1, fmt='(a,f11.6)')    'Error:                                                      RMSE(tau) = ',tau_est_RMSE_work
            end if
            if (opt2 .eq. 2 .and. pickg .eq. 'y') then
               if (gamma_est_work .ge. 0.0_sp .and. gamma_est_work .lt. 10.0_sp) then
                  write (unit=1, fmt='(a,f8.6)')  'Tail index estimate:                                           gamma  = ',gamma_est_work
               else if (gamma_est_work .ge. 10.0_sp .and. gamma_est_work .lt. 100.0_sp) then
                  write (unit=1, fmt='(a,f9.6)')  'Tail index estimate:                                           gamma  = ',gamma_est_work
               else if (gamma_est_work .ge. 100.0_sp .and. gamma_est_work .lt. 1000.0_sp) then
                  write (unit=1, fmt='(a,f10.6)') 'Tail index estimate:                                           gamma  = ',gamma_est_work
               else
                  write (unit=1, fmt='(a,f11.6)') 'Tail index estimate:                                           gamma  = ',gamma_est_work
               end if
               if (gamma_est_RMSE_work .ge. 0.0_sp .and. gamma_est_RMSE_work .lt. 10.0_sp) then
                  write (unit=1, fmt='(a,f8.6)')  'Error:                                                    RMSE(gamma) = ',gamma_est_RMSE_work
               else if (gamma_est_RMSE_work .ge. 10.0_sp .and. gamma_est_RMSE_work .lt. 100.0_sp) then
                  write (unit=1, fmt='(a,f9.6)')  'Error:                                                    RMSE(gamma) = ',gamma_est_RMSE_work
               else if (gamma_est_RMSE_work .ge. 100.0_sp .and. gamma_est_RMSE_work .lt. 1000.0_sp) then
                  write (unit=1, fmt='(a,f10.6)') 'Error:                                                    RMSE(gamma) = ',gamma_est_RMSE_work
               else
                  write (unit=1, fmt='(a,f11.6)') 'Error:                                                    RMSE(gamma) = ',gamma_est_RMSE_work
               end if
            else
               if (alpha_est_work .ge. 0.0_sp .and. alpha_est_work .lt. 10.0_sp) then
                  write (unit=1, fmt='(a,f8.6)')  'Tail index estimate:                                           alpha  = ',alpha_est_work
               else if (alpha_est_work .ge. 10.0_sp .and. alpha_est_work .lt. 100.0_sp) then
                  write (unit=1, fmt='(a,f9.6)')  'Tail index estimate:                                           alpha  = ',alpha_est_work
               else if (alpha_est_work .ge. 100.0_sp .and. alpha_est_work .lt. 1000.0_sp) then
                  write (unit=1, fmt='(a,f10.6)') 'Tail index estimate:                                           alpha  = ',alpha_est_work
               else
                  write (unit=1, fmt='(a,f11.6)') 'Tail index estimate:                                           alpha  = ',alpha_est_work
               end if
               if (alpha_est_RMSE_work .ge. 0.0_sp .and. alpha_est_RMSE_work .lt. 10.0_sp) then
                  write (unit=1, fmt='(a,f8.6)')  'Error:                                                    RMSE(alpha) = ',alpha_est_RMSE_work
               else if (alpha_est_RMSE_work .ge. 10.0_sp .and. alpha_est_RMSE_work .lt. 100.0_sp) then
                  write (unit=1, fmt='(a,f9.6)')  'Error:                                                    RMSE(alpha) = ',alpha_est_RMSE_work
               else if (alpha_est_RMSE_work .ge. 100.0_sp .and. alpha_est_RMSE_work .lt. 1000.0_sp) then
                  write (unit=1, fmt='(a,f10.6)') 'Error:                                                    RMSE(alpha) = ',alpha_est_RMSE_work
               else
                  write (unit=1, fmt='(a,f11.6)') 'Error:                                                    RMSE(alpha) = ',alpha_est_RMSE_work
               end if                
            end if
            if (k_opt .lt. 1) then
               write (unit=1, fmt='(a,i4)')       'Optimal order:                                                  k_opt = ',k_opt
            else if (k_opt .ge. 1     .and. k_opt .lt. 10) then
               write (unit=1, fmt='(a,i1)')       'Optimal order:                                                  k_opt = ',k_opt
            else if (k_opt .ge. 10    .and. k_opt .lt. 100) then
               write (unit=1, fmt='(a,i2)')       'Optimal order:                                                  k_opt = ',k_opt
            else if (k_opt .ge. 100   .and. k_opt .lt. 1000) then
               write (unit=1, fmt='(a,i3)')       'Optimal order:                                                  k_opt = ',k_opt
            else if (k_opt .ge. 1000  .and. k_opt .lt. 10000) then
               write (unit=1, fmt='(a,i4)')       'Optimal order:                                                  k_opt = ',k_opt
            else if (k_opt .ge. 10000 .and. k_opt .lt. 100000) then
               write (unit=1, fmt='(a,i5)')       'Optimal order:                                                  k_opt = ',k_opt
            else if (k_opt .ge. 100000) then
               write (unit=1, fmt='(a,i6)')       'Optimal order:                                                  k_opt = ',k_opt
            end if
            write (unit=1, fmt='(a)')             ' '
            write (unit=1, fmt='(a)')             'Notes: (1) alpha = -999.0 means that no estimation is possible since for all k,'
            write (unit=1, fmt='(a)')             '           alpha(k) is outside of the interval [0; 2] (exclusive of the endpoints).'
            write (unit=1, fmt='(a)')             '           Suggested options for re-analysis:'
            write (unit=1, fmt='(a)')             '           try other preprocessing types or an increased data size.'
            write (unit=1, fmt='(a)')             '       (2) RMSE(k) is the measure for order selection.'
            write (unit=1, fmt='(a)')             '       (3) RMSE(k) = -999.0 means that this k value has not been evaluated.'
            write (unit=1, fmt='(a)')             ' '
            if (opt2 .eq. 2 .and. pickg .eq. 'y') then
               write (unit=1, fmt='(10(a))')                                                                           &
                     ' t_original    ',                                                                                &
                     ' X_original    ',                                                                                &
                     ' t             ',                                                                                &
                     ' X             ',                                                                                &
                     ' X_sorted      ',                                                                                &
                     '     k',                                                                                         &
                     '  gamma(k)     ',                                                                                &
                     '  RMSE(k)      ',                                                                                &
                     '  gamma-RMSE(k)',                                                                                &
                     '  gamma+RMSE(k)'
            else
               write (unit=1, fmt='(10(a))')                                                                           &
                     ' t_original    ',                                                                                &
                     ' X_original    ',                                                                                &
                     ' t             ',                                                                                &
                     ' X             ',                                                                                &
                     ' X_sorted      ',                                                                                &
                     '     k',                                                                                         &
                     '  alpha(k)     ',                                                                                &
                     '  RMSE(k)      ',                                                                                &
                     '  alpha-RMSE(k)',                                                                                &
                     '  alpha+RMSE(k)'
            end if
            allocate(Yordered(Nobs))
            Yordered=Y
            call sort(Yordered,Y,Nobs,1,Nobs)
            Yordered(1:Nobs)=Yordered(Nobs:1:-1)
            do i=1,K_result-1
               if (alpha_sequence_result(i) .eq. -999.0_sp .or. RMSE_result(i) .eq. -999.0_sp) then
                  alpha_output_low=-999.0_sp
                  alpha_output_upp=-999.0_sp
               else
                  alpha_output_low=alpha_sequence_result(i)-RMSE_result(i)
                  alpha_output_upp=alpha_sequence_result(i)+RMSE_result(i)
               end if
               write (unit=1, fmt='(5(es14.6e3,1x),i6,4(1x,es14.6e3))')                                                &
                     torig(i),Yorig(i),t(i),Y(i),Yordered(i),i,alpha_sequence_result(i),RMSE_result(i),alpha_output_low,alpha_output_upp
            end do
            if (allocated(alpha_sequence)) deallocate(alpha_sequence)
            if (allocated(RMSE)) deallocate(RMSE)
            if (K_result .ge. 1) then
               do i=K_result,Nobs
                  write (unit=1, fmt='(5(es14.6e3,1x))')             torig(i),Yorig(i),t(i),Y(i),Yordered(i)
               end do
            end if
            close (unit=1, status='keep')
            if (allocated(Yordered)) deallocate(Yordered)
            print *
            print '(a,a)',' Full results written into file:                                          ',outfile
!
            open (unit=1, file=outfilealpha, status='unknown', form='formatted', action='write')    ! ht-d-stats
            if (opt2 .eq. 2 .and. pickg .eq. 'y') then
               write (unit=1, fmt='(f11.6)') one/alpha_est_work
            else
               write (unit=1, fmt='(f11.6)') alpha_est_work
            end if
            close (unit=1, status='keep')
            open (unit=1, file=outfilekopt, status='unknown', form='formatted', action='write')     ! ht-d-stats
            write (unit=1, fmt='(i6)') k_opt
            close (unit=1, status='keep')
            open (unit=1, file=outfiledata, status='unknown', form='formatted', action='write')     ! ht-d-stats
            do i=1,Nobs
               write (unit=1, fmt='(2(es14.6e3,1x))') t(i),Y(i)
            end do
            close (unit=1, status='keep')
!
! 9.4       Output file, one line of reduced content (modus 'estimate')
!           ===========================================================
!
!           Note: If the input series has too few positive values, then it may happen that the estimation is stopped
!                 (Section 4), without any output being written into outfile_red. This situation may occur in 'external'
!                 MC experiments, where random input data are generated (modus 'generate').
!
            if (modus .eq. 'estimate') then
               outfile_red=trim(adjustl(SUBR_EST_OUT_outputfile2))
               open (unit=2, file=outfile_red, status='unknown', form='formatted', action='read', iostat = iocheck)
               if (iocheck .eq. 0 ) then
                  read (unit=2, fmt='(a)') dummyc1
               else
                  dummyc1=' '
               end if
               close (unit=2, status='keep')
               if (trim(adjustl(dummyc1)) .eq. '') then ! if file does not exist, then open new file and write header
                  open (unit=2, file=outfile_red, status='unknown', form='formatted', action='write')
                  write (unit=2, fmt='(7(1x,a))') ' k_opt',                                                             &
                                                  ' tau          ',                                                     &
                                                  ' RMSE(tau)    ',                                                     &
                                                  ' alpha        ',                                                     &
                                                  ' RMSE(alpha)  ',                                                     &
                                                  ' gamma        ',                                                     &
                                                  ' RMSE(gamma)  '
                  close (unit=2, status='keep')
               end if
               open (unit=2, file=outfile_red, status='unknown', form='formatted', action='write', position='append')
               write (unit=2, fmt='(1x,i6,6(1x,es14.6e3))') k_opt,   tau_est_work,   tau_est_RMSE_work,                 &
                                                                   alpha_est_work, alpha_est_RMSE_work,                 &
                                                                   gamma_est_work, gamma_est_RMSE_work
               close (unit=2, status='keep')
            end if
!
! 9.5       Output flag (end)
!           =================
!
            end if
!
! 10.       Final stuff
!           ===========
!            
            if (modus .eq. 'per-hand' .and. outflag .eq. 'y') then
               print *
               print '(a)',  ' PRESS ENTER TO CONTINUE'
               read (*,*)
            end if
        end subroutine est_out
        
        !-------------------------------
        ! EXP[N(0,1)^2/b^2] distribution
        !-------------------------------

        subroutine expn01(alpha,x,n,rng3,rng4)
        use nrtype
        use rng
        implicit none
        real(sp), intent(in)  :: alpha                              ! tail index
        integer, intent(in)  :: n                                   ! data size
        real(sp), dimension(n), intent(out) :: x                    ! variable
        type(rng_t), dimension(1), intent(inout) :: rng3            ! seeded random number generator
        type(rng_t), dimension(1), intent(inout) :: rng4            ! seeded random number generator
!
!       Distribution generated using a standard normal distribution (analogously to the lognormal distribution).
!       It is not a stable distribution, and its heavy tail index parameter can be selected via
!       alpha = b^2/2.
!       (For the sake of maintaining the original notation of our notes, we keep usage of the parameter b,
!       although we do not need it here for the simulations.)
!       The probability density function can be expressed in closed analytical form.
!       The mean of this distribution exists for b > SQRT(2) or alpha > 1.
!       The variance exists for b > 2 or alpha > 2.
!       For more details, see the reference for this software
!       (Mudelsee and Bermejo), which is cited in the header of this source code.
!
        integer :: i
        integer :: iseed=1                                          ! dummy integer for random number generator
        real(sp), parameter :: two=2.0_sp
        do i=1,n
           call gaussian(x(i),rng3(iseed),rng4(iseed))
           x(i)=exp(x(i)**two/(two*alpha))
        end do
        end subroutine expn01

        !-------------------------------
        ! Gaussian standard distribution
        !-------------------------------

        subroutine gaussian(z,rng3,rng4)
        use nrtype
        use rng
        implicit none
        real(sp), intent(out) :: z
        type(rng_t), dimension(1), intent(inout) :: rng3            ! seeded random number generator
        type(rng_t), dimension(1), intent(inout) :: rng4            ! seeded random number generator
!
!       Gaussian standard distribution using the transformation after Box and Muller (1958).
!       Box GEP, Muller ME (1958) Ann Math Stat 29:610.
!
        real(sp) :: z1
        real(sp) :: z2
        real(sp) :: zsq
        real(sp), parameter :: zero=0.0_sp
        real(sp), parameter ::  one=1.0_sp
        real(sp), parameter ::  two=2.0_sp
        integer :: iseed=1                                          ! dummy integer for random number generator
        do
           z1=mzran(rng3(iseed))
           z2=mzran(rng4(iseed))
           z1=two*z1-one
           z2=two*z2-one
           zsq=z1**two+z2**two
           if (zsq > zero .and. zsq < one) exit
        end do
        zsq=sqrt(-two*log(zsq)/zsq)
        z=z1*zsq
        end subroutine gaussian

        !---------------------------
        ! ARTIFICIAL DATA GENERATION
        !---------------------------

        subroutine generate(torig,Norig,rng3,rng4)
            use nrtype
            use cfgseedparameters, only: modus,SUBR_GENERATE_tau,SUBR_GENERATE_alpha,SUBR_GENERATE_shape,SUBR_GENERATE_artfile
            use datafilename
            use Hill_Pick_testNpos_alpha
            use rng
            implicit none
            integer, intent(in)  :: Norig                               ! number of observations (original series)
            real(sp), dimension(Norig), intent(in) :: torig             ! time (original series)
            type(rng_t), dimension(1), intent(inout) :: rng3            ! seeded random number generator
            type(rng_t), dimension(1), intent(inout) :: rng4            ! seeded random number generator
!
!           Generation of artificial data.
!
!           Included is a test for suitable output:
!           (1) at least minNpositive positive values (after mean subtraction),
!           (2) at least for one order value, the tail index alpha (Hill, Pickands) is between 0 and 2.
!
            character(len=1) :: answer1                                 ! answer
            character(len=1) :: answer2                                 ! answer
            character(len=16) :: artfile='hta.dat'                      ! output file name
            character(len=1) :: flagtgen='-'                            ! flag for testing whether alpha is between 0 and 2
            integer, parameter :: tmp_opt2=1                            ! temporal integer for opt2
                                                                        ! temp_opt2 = 1: Hill
            integer :: i                                                ! counter
            integer :: iseed=1                                          ! dummy integer for random number generator
            integer :: n                                                ! data size
            real(sp) :: alpha                                           ! tail index
            real(sp) :: phi                                             ! (equivalent) autocorrelation coefficient
            real(sp) :: tau                                             ! persistence time
            real(sp), dimension(:), allocatable :: t                    ! time (artificial series)
            real(sp), dimension(:), allocatable :: x                    ! data (artificial series)
            real(sp), dimension(:), allocatable :: w                    ! generated process based on the original Y
            real(sp), parameter :: zero=0.0_sp
            real(sp), parameter ::  one=1.0_sp
            real(sp) :: z1=-999.0_sp                                    ! variable, Gaussian distribution
            real(sp) :: z2=-999.0_sp                                    ! variable, Gaussian distribution
!
! 1.        Modus 'per-hand'
!           ================
!
            if (modus .eq. 'per-hand') then
               print *
               print '(a)',                 ' Generation of artificial time series'
               print *
               print '(a)',                 ' 1. Time grid: original or evenly spaced'
               print '(a)',                 ' 2. Data size (even spacing): selectable'
               print '(a)',                 ' 3. Serial dependence: AR(1)'
               print '(a)',                 ' 4. Distributional shape and heavy tail parameter: selectable'
               print *
               print '(a)',                 ' 1. Time grid'
               print '(a,i6,a)',            '    Overtake original time values of size ',Norig,' ?              [Y/N]'
               read *, answer1
               do
                  if (answer1 .eq. 'Y' .or. answer1 .eq. 'y') then
                     n=Norig
                     allocate(t(n))
                     allocate(x(n))
                     t=torig
                     print '(a)',           ' 2. Data size'
                     print '(a)',           '    Original data size overtaken'
                     exit
                  else if (answer1 .eq. 'N' .or. answer1 .eq. 'n') then
                     print '(a)',           ' 2. Data size for even spacing unity time grid:               [INPUT]'
                     print '(a,i6,a,i2,a)', '    (maximum: ',Nobsmax,', minimum: ',Nobsmin,')'
                     read *, n
                     if (n .ge. Nobsmin .and. n .le. Nobsmax) then
                        allocate(t(n))
                        allocate(x(n))
                        do i=1,n
                           t(i)=1.0_sp*i
                        end do
                        exit
                     end if
                  end if
               end do
               print *
               print '(a)',                 ' 3. Serial dependence'
               print '(a)',                 '    Prescribe AR(1) persistence time (>= 0)                   [INPUT]'
               read *, tau
               print *
               print '(a)',                 ' 4. Distributional shape'
               print '(a)',                 '    Cauchy (alpha = 1)                                            [c]'
               print '(a)',                 '    EXP[N(0,1)^2/b^2] (alpha = b^2/2> 0 selectable)               [e]'
               print '(a)',                 '    Gaussian standard N(0,1) (alpha = 2)                          [g]'
               print '(a)',                 '    Stable (2 > alpha > 0 selectable, alpha > 0.4 recommended)    [s]'
               read *, answer2
               do
                  if (answer2 .eq. 'C' .or. answer2 .eq. 'E' .or. answer2 .eq. 'G' .or. answer2 .eq. 'S' .or.             &
                      answer2 .eq. 'c' .or. answer2 .eq. 'e' .or. answer2 .eq. 'g' .or. answer2 .eq. 's') exit
               end do
               if (answer2 .eq. 'C') answer2='c'
               if (answer2 .eq. 'E') answer2='e'
               if (answer2 .eq. 'G') answer2='g'
               if (answer2 .eq. 'S') answer2='s'
!
! 2.        Modus 'generate'
!           ================
!
            else if (modus .eq. 'generate') then
!
! 2.1       Check data size
!           ===============
!
               n=Norig
               if (n .gt. Nobsmax) then
                  print '(a,i6,a)', ' MAXIMUM NUMBER OF OBSERVATIONS (',Nobsmax,') EXCEEDED'
                  print *
                  print '(a)',      ' Program terminates.'
                  stop
               end if
               if (n .lt. Nobsmin) then
                  print '(a,i6,a)', ' MINIMUM NUMBER OF OBSERVATIONS (',Nobsmin,') MISSED'
                  print *
                  print '(a)',      ' Program terminates.'
                  stop
               end if
!
! 2.2       Time grid
!           =========
!
               allocate(t(n))
               allocate(x(n))
               do i=1,n
                  t(i)=1.0_sp*i
               end do
!
! 2.3       Persistence time, alpha, shape
!           ==============================
!
               tau=SUBR_GENERATE_tau
               alpha=SUBR_GENERATE_alpha
               answer2=SUBR_GENERATE_shape
            end if
!
! 3.        Start test loop for suitable output
!           ===================================
!
testnp6:    do itestnp=1,ntestnp
!
!
! 4.           Generate innovations
!              ====================
!
               if (answer2 .eq. 'c') then
                  do i=1,n
                     call cauchy(x(i),rng3(iseed))
                  end do
               else if (answer2 .eq. 'e') then
                  if (modus .eq. 'per-hand') then
                     print '(a)',              ' 4. Distributional shape: EXP[N(0,1)^2/b^2] (alpha = b^2/2)'
                     print '(a)',              '    Prescribe alpha                                           [INPUT]'
                     read *, alpha
                  end if
                  call expn01(alpha,x,n,rng3(iseed),rng4(iseed))
               else if (answer2 .eq. 'g') then
                  do i=1,n
                     call gaussian(x(i),rng3(iseed),rng4(iseed))
                  end do
               else if (answer2 .eq. 's') then
                  if (modus .eq. 'per-hand') then
                     print *
                     print '(a)',              ' 4. Distributional shape: Stable'
                     print '(a)',              '    Prescribe 2 > alpha > 0                                   [INPUT]'
                     read *, alpha
                  end if
                  call stable(alpha,x,n,rng3(iseed),rng4(iseed))
               end if
!
! 5.           Generate AR(1) time series
!              ==========================
!
               do i=2,n
                  if (tau .eq. zero) then
                     phi=zero
                  else
                     phi=exp(-(t(i)-t(i-1))/tau)
                  end if
                  x(i)=phi*x(i-1)+(one-phi*phi)*x(i)               ! generating data from an unevenly spaced, heavy-tailed AR(1) noise process;
                                                                   ! factor (one-phi*phi) ensures strict stationarity
                                                                   ! (Mudelsee 2014 Climate Time Series Analysis, 2nd edn., Springer, Cham: Ch. 2 therein),
                                                                   ! which may be relevant for finite-variance processes
               end do
!
! 6.           Test for suitable output
!              ========================
!
               ! at least minNpositive positive values of mean-subtracted Ysim are required for Hill estimator
               testNpositive=count((x-sum(x)/n) > 0.0_sp)
               ! Test (over the full order search range) whether at least for one order value
               ! the tail index alpha (on the simulated data) is between 0 and 2.
               call test_alpha_initial(x,n,tmp_opt2,flagtgen)
               if (flagtgen .eq. '+' .and. .not. (tmp_opt2 .eq. 1 .and. testNpositive .lt. minNpositive)) exit testnp6
           end do testnp6
!
! 7.        Output
!           ======
!
            if (modus .eq. 'per-hand') then
               if (inputfile .eq. artfile) then
                  print *
                  print '(a)',  ' ALERT:'
                  print *
                  print '(a,a)',' Current input data file: ',trim(adjustl(inputfile))
                  print '(a,a)',' Planned output data file: ',artfile
                  print *
                  print '(a)',  ' You may wish to rename and save the current input file now since it will be overwritten.'
                  print *
                  print '(a)',  ' PRESS ENTER TO CONTINUE'
                  read (*,*)
               end if
            elseif (modus .eq. 'generate') then
               artfile=trim(adjustl(SUBR_GENERATE_artfile))
            end if
            open (unit=1, file=artfile, status='unknown', form='formatted', action='write')
            do i=1,n
               write (unit=1, fmt='(2(1x,es14.6e3))') t(i), x(i)
            end do
            close (unit=1, status='keep')
            if (modus .eq. 'per-hand') then
               print *
               print '(a,a)',' Artificial time series written into file: ',artfile
               print *
               print '(a)',  ' For analysing the new output data file within this software run, you have to load it (main menu).'
               print *
               print '(a)',  ' PRESS ENTER TO CONTINUE'
               read (*,*)
            end if
!
! 8.        Exit
!           ====
!
            deallocate(t)
            deallocate(x)
        end subroutine generate

        !--------------------------------------------------
        ! ARTIFICIAL DATA GENERATION FOR THE MC SIMULATIONS
        !--------------------------------------------------

        subroutine generateMC(t,x,n,tau,alpha,gamma,opt2,pickg,rng3,rng4)
            use nrtype
            use rng
            implicit none
            integer, intent(in)  :: n                                   ! data size
            real(sp), dimension(n), intent(in) :: t                     ! time
            real(sp), dimension(n), intent(out) :: x                    ! variable
            real(sp), intent(in) :: tau                                 ! persistence time
            real(sp), intent(in) :: alpha                               ! tail index
            real(sp), intent(in) :: gamma                               ! tail index (gamma=1/alpha)
            integer, intent(in) :: opt2                                 ! opt2 = 1: Hill
                                                                        ! opt2 = 2: Pickands
            character(len=1), intent(in) :: pickg                       ! pickg = 'y': Pickands estimator calculates gamma
                                                                        ! pickg = 'n': Pickands estimator calculates alpha
            type(rng_t), dimension(1), intent(inout) :: rng3            ! seeded random number generator
            type(rng_t), dimension(1), intent(inout) :: rng4            ! seeded random number generator
!
!           Generation of artificial data for the MC simulations.
!
            integer :: i                                                ! counter
            integer :: iseed=1                                          ! dummy integer for random number generator
            real(sp) :: alphasim                                        ! tail index (alpha) used for the MC simulations
            real(sp) :: phi                                             ! (equivalent) autocorrelation coefficient
            real(sp), parameter :: zero=0.0_sp
            real(sp), parameter ::  one=1.0_sp
!
! 1.        Determine tail index
!           ====================
!
            if (opt2 .eq. 2 .and. pickg .eq. 'y') then
               alphasim=one/gamma
            else
               alphasim=alpha
            end if
!
! 2.        Generate innovations
!           ====================
!


            ! CHECK Version 2.0
            !
            !
            !   call stable(alphasim,x,n,rng3(iseed),rng4(iseed))
            call expn01(alphasim,x,n,rng3(iseed),rng4(iseed))
            !
            !
            !
!
! 3.        Generate AR(1) time series
!           ==========================
!
            do i=2,n
               if (tau .eq. zero) then
                  phi=zero
               else
                  phi=exp(-(t(i)-t(i-1))/tau)
               end if
               x(i)=phi*x(i-1)+(one-phi*phi)*x(i)       ! generating data from an unevenly spaced, heavy-tailed AR(1) noise process;
                                                        ! factor (one-phi*phi) ensures strict stationarity
                                                        ! (Mudelsee 2014 Climate Time Series Analysis, 2nd edn., Springer, Cham: Ch. 2 therein),
                                                        ! which may be relevant for finite-variance processes
            end do
        end subroutine generateMC

        !---------------
        ! HILL ESTIMATOR
        !----------------

        subroutine hill(Y,N,alpha_sequence,Npositive)
            use nrtype
            implicit none
            integer, intent(in) :: N                                        ! input data size
            real(sp), dimension(N), intent(in) :: Y                         ! input data
            integer, intent(in) :: Npositive                                ! number of positive input data values, determines output data size
            real(sp), dimension(Npositive-1), intent(out) :: alpha_sequence ! output data
!
!           Hill estimator of tail index.
!
!           We follow Resnick (2007 Heavy-Tail Phenomena: Probabilistic and Statistical Modeling, Springer, New York: pp. 74, 364 therein)
!           and use only the positive observations.
!
!           Note that shifting Y and, therefore, modifying the implementation of the Hill estimator is possible
!           during execution of the software at preprocessing.
!
!           Note also a safeguard (tiny) against the case that the Npositive values are almost identical,
!           1/alpha is almost equal to zero and alpha is almost infinity.
!
            real(sp), allocatable, dimension(:) :: Yordered                 ! auxiliary array
            real(sp), allocatable, dimension(:) :: gamma1                   ! auxiliary array
            real(sp), allocatable, dimension(:) :: oneoverk                 ! auxiliary array
                                                                            ! declaring Yordered and oneoverk allocatable saves RAM
            integer :: i                                                    ! counter
            real(sp), parameter :: one=1.0_sp
            real(sp), parameter :: tiny=1.0e-009_sp                         ! tiny value (oriented on the machine precision)
!
            allocate(Yordered(N))
            Yordered=Y
            call sort(Yordered,Y,N,1,N)
            Yordered(1:Npositive)=log(Yordered(N:N-Npositive+1:-1))         ! Natural logarithm
            allocate(gamma1(Npositive-1))
            allocate(oneoverk(Npositive-1))
            gamma1(1)=Yordered(1)
            oneoverk(1)=one
            do i=2,Npositive-1
               gamma1(i)=gamma1(i-1)+Yordered(i)
               oneoverk(i)=one/real(i)
            end do
            do i=1,Npositive-1
               alpha_sequence(i)=oneoverk(i)*gamma1(i)-Yordered(i+1)
            end do
            where (alpha_sequence .lt. tiny) alpha_sequence=tiny
            alpha_sequence=one/alpha_sequence
            deallocate(oneoverk)
            deallocate(gamma1)
            deallocate(Yordered)
        end subroutine hill

        !-----------------------------------------------------------------------
        ! Least-squares function for subroutine minls for persistence estimation
        !-----------------------------------------------------------------------

        double precision function ls(a,t,x,n)
            implicit none
            integer n
            double precision t(1:n),x(1:n)
            double precision a
            integer i
            ls=0.0d0
            do i=2,n
               ls=ls+(x(i)-x(i-1)*dsign(1.0d0,a)* dabs(a)**(t(i)-t(i-1)))**2.0d0
            end do
            return
        end function ls

        !-----------------------
        ! MENU FOR PREPROCESSING
        !-----------------------

        integer function menu_data()
            implicit none
            integer :: opt
            do
                call clearscreen
                print '(a)', ' Select preprocessing type                                    [ENTER A NUMBER]'
                print *
                print '(a)', ' 1. Reverse time direction'
                print '(a)', ' 2. Transform the data otherwise'
                print '(a)', ' 3. Main menu'
                read *, opt
                call clearscreen
                if (opt>=1 .and. opt<=3) exit
                print '(a)', ' PLEASE TYPE IN A NUMBER BETWEEN 1 AND 3'
            end do
            menu_data=opt
        end function menu_data

        !------------------------------
        ! MENU FOR TAIL INDEX ESTIMATOR
        !------------------------------

        integer function menu_estimator()
            implicit none
            integer :: opt
            do
                call clearscreen
                print '(a)', ' Select tail index estimator                                  [ENTER A NUMBER]'
                print *
                print '(a)', ' 1. Hill estimator'
                print '(a)', ' 2. Pickands estimator'
                print '(a)', ' 3. Main menu'
                read *, opt
                call clearscreen
                if (opt>=1 .and. opt<=3) exit
                print '(a)', ' PLEASE TYPE IN A NUMBER BETWEEN 1 AND 3'
            end do
            menu_estimator=opt
        end function menu_estimator

        !----------------------------------
        ! MENU FOR NUMBER OF MC SIMULATIONS
        !----------------------------------

        integer function menu_nsim()
            implicit none
            integer :: i
            integer, parameter :: nsimmin=10 ! minimum number of MC simulations
            do
                call clearscreen
                print '(a)', ' Number of Monte Carlo (MC) simulations for RMSE calculations          [INPUT]'
                print *
                print '(a,i2)', ' Minimum:        ',nsimmin
                print '(a)',    ' Recommendation: at least 50'
                print '(a)',    ' Negative value: no MC simulations'
                read *, i
                call clearscreen
                if (i .lt. 0 .or. i .ge. nsimmin) exit
            end do
            if (i .lt. 0) i=-999
            menu_nsim=i
        end function menu_nsim

        !------------------------------------------
        ! MENU FOR TAU ESTIMATION IN MC SIMULATIONS
        !------------------------------------------

        subroutine menu_MC_tauflag(tauflag)
            implicit none
            character(len=1), intent(out) :: tauflag
            do
                call clearscreen
                print '(a)', ' MC simulations also for RMSE(tau) calculation?                          [Y/N]'
                read '(a1)', tauflag
                call clearscreen
                if (tauflag .eq. 'Y' .or. tauflag .eq. 'y' .or. tauflag .eq. 'N' .or. tauflag .eq. 'n') exit
            end do
        end subroutine menu_MC_tauflag

        !-------------------------
        ! MENU FOR ORDER SELECTION
        !-------------------------

        subroutine menu_order_selection(Nobs)
            use qbfparameters
            implicit none
            integer, intent(in) :: Nobs
            character(len=1) :: answer
            call clearscreen
            do
                print '(a)', ' Order selector                                               [ENTER A NUMBER]'
                print *
                print '(a)', ' 1. Brute force'
                print '(a)', ' 2. Quasi-brute force'
                read *, optBF
                if (optBF>=1 .and. optBF<=2) exit
                print '(a)', ' PLEASE TYPE IN A NUMBER BETWEEN 1 AND 2'
            end do
            if (optBF .eq. 2) then
               call clearscreen
               print '(a)',        ' Current quasi-brute force parameter values:'
               print *
               print '(a,i6)',     ' Length of search-steps used in Step 1:             L_k = ',L_k
               print '(a,i6)',     ' A good choice for not too large data size is:      L_k = ',5
               print *
               print '(a,i6)',     ' Note: L_k should be clearly less than the data size      ',Nobs
               print *
               print '(a,i6)',     ' Percentage of best k values used in Step 2:      Pmink = ',Pmink
               print '(a,i6)',     ' A good choice for not too large data size is:    Pmink = ',5
               print *
               do
                   answer='n'
                   print '(a)',    ' Do you want to change L_k or Pmink?                          [Y/N]'
                   read *, answer
                   if (answer=='Y' .or. answer=='y') then
                      print '(a)', ' Set L_k, Pmink (two values comma-separated)                  [INPUT]'
                      read *, L_k,Pmink
                      if (L_k .ge. 1 .and. L_k .le. Nobs .and. Pmink .ge. 1 .and. Pmink .le. 100) exit
                   else
                      exit
                   end if
               end do
            end if
        end subroutine menu_order_selection

        !---------------------
        ! TRANSFORM VARIABLE Y
        !---------------------
        
        subroutine menu_transform(Y,t,Nobs,Ytrans,ttrans,Ntrans)
            implicit none
            integer, intent(in) :: Nobs
            real(sp), dimension(1:Nobs), intent(in)  :: Y
            real(sp), dimension(1:Nobs), intent(in)  :: t
            real(sp), dimension(1:Nobs), intent(out) :: Ytrans
            real(sp), dimension(1:Nobs), intent(out) :: ttrans
            integer, intent(out) :: Ntrans
            integer :: i,j
            integer :: opt
            real(sp) :: a,b,z
            real(sp), parameter :: one=1.0_sp
            do
                call clearscreen
                print '(a)', ' Original                        time series: Torig(1:Norig), Xorig(1:Norig)'
                print '(a)', ' Current (possibly preprocessed) time series: T(1:N),         X(1:N)'
                print *
                print '(a)', ' Select transformation for X                                  [ENTER A NUMBER]'
                print *
                print '(a)', ' 1. Natural log transformation'
                print '(a)', '    X = log(X)'
                print *
                print '(a)', ' 2. Linear transformation'
                print '(a)', '    X = a + b * X'
                print *
                print '(a)', ' 3. First differences'
                print '(a)', '    X(1:N-1) = X(2:N) - X(1:N-1)'
                print '(a)', '    Note: data size reduced by one'
                print *
                print '(a)', ' 4. Returns'
                print '(a)', '    X(1:N-1) = [X(2:N) - X(1:N-1)] / [X(1:N-1)]'
                print '(a)', '    Note: data size reduced by one'
                print *
                print '(a)', ' 5. Natural log returns'
                print '(a)', '    X(1:N-1) = log[X(2:N)] - log[X(1:N-1)]'
                print '(a)', '    Note: data size reduced by one'
                print *
                print '(a)', ' 6. Retain only values above a selectable threshold'
                print '(a)', '    X(1:N) > [threshold]'
                print '(a)', '    Note: data size reduced'
                print *
                print '(a)', ' 7. Previous menu'
                read *, opt
                call clearscreen
                if (opt==1) then
                    Ytrans=log(Y)
                    ttrans=t
                    Ntrans=Nobs
                    exit
                else if (opt==2) then
                    print '(a)', ' X = a + b * X'
                    print *
                    print '(a)', ' Type in coefficient a:'
                    read *, a
                    a=a*one
                    print '(a)', ' Type in coefficient b:'
                    read *, b
                    b=b*one
                    Ytrans=a+b*Y
                    ttrans=t
                    Ntrans=Nobs
                    exit
                else if (opt==3) then
                    Ytrans(1:Nobs-1)=Y(2:Nobs)-Y(1:Nobs-1)
                    ttrans(1:Nobs-1)=t(1:Nobs-1)
                    Ntrans=Nobs-1
                    exit
                else if (opt==4) then
                    Ytrans(1:Nobs-1)=(Y(2:Nobs)-Y(1:Nobs-1))/Y(1:Nobs-1)
                    ttrans(1:Nobs-1)=t(1:Nobs-1)
                    Ntrans=Nobs-1
                    exit
                else if (opt==5) then
                    Ytrans(1:Nobs-1)=log(Y(2:Nobs))-log(Y(1:Nobs-1))
                    ttrans(1:Nobs-1)=t(1:Nobs-1)
                    Ntrans=Nobs-1
                    exit
                else if (opt==6) then
                    print '(a)', ' Type in threshold value:'
                    read *, z
                    z=z*one
                    j=0
                    Ntrans=0
                    do i=1,Nobs
                       if (Y(i) .gt. z) then
                          j=j+1
                          Ytrans(j)=Y(i)
                          ttrans(j)=t(i)
                          Ntrans=Ntrans+1
                       end if
                    end do
                    if (Nobs .lt. 100) then
                       print '(a,i2)',                 ' Data size (before preprocessing):                      ',Nobs
                    else if (Nobs .ge. 100 .and. Nobs .lt. 1000) then
                       print '(a,i3)',                 ' Data size (before preprocessing):                      ',Nobs
                    else if (Nobs .ge. 1000 .and. Nobs .lt. 10000) then
                       print '(a,i4)',                 ' Data size (before preprocessing):                      ',Nobs
                    else if (Nobs .ge. 10000 .and. Nobs .lt. 100000) then
                       print '(a,i5)',                 ' Data size (before preprocessing):                      ',Nobs
                    else if (Nobs .ge. 100000) then
                       print '(a,i6)',                 ' Data size (before preprocessing):                      ',Nobs
                    end if
                    if (Ntrans .lt. 100) then
                       print '(a,i2)',                 ' Data size (after preprocessing):                       ',Ntrans
                    else if (Ntrans .ge. 100 .and. Ntrans .lt. 1000) then
                       print '(a,i3)',                 ' Data size (after preprocessing):                       ',Ntrans
                    else if (Ntrans .ge. 1000 .and. Ntrans .lt. 10000) then
                       print '(a,i4)',                 ' Data size (after preprocessing):                       ',Ntrans
                    else if (Ntrans .ge. 10000 .and. Ntrans .lt. 100000) then
                       print '(a,i5)',                 ' Data size (after preprocessing):                       ',Ntrans
                    else if (Ntrans .ge. 100000) then
                       print '(a,i6)',                 ' Data size (after preprocessing):                       ',Ntrans
                    end if
                    print *
                    print '(a)', ' PRESS ENTER TO CONTINUE'
                    read (*,*)
                    exit
                else if (opt==7) then
                    exit
                end if
                print '(a)', ' PLEASE TYPE IN A NUMBER BETWEEN 1 AND 7'
            end do
            print *
        end subroutine menu_transform

        !---------------------------------------------------------------------
        ! Minimization of least-squares function ls for persistence estimation
        !---------------------------------------------------------------------

        subroutine minls(n, t, x, amin, nmu_)
            implicit none

            double precision, parameter :: a_ar1 = 0.367879441d0 ! 1/e
            double precision, parameter :: tol = 3.0d-8          ! Brent's search, precision
            double precision, parameter :: tol2 = 1.0d-6         ! multiple solutions, precision
            integer n
            double precision t(1:n),x(1:n)
            double precision amin
            integer nmu_
            double precision dum1,dum2,dum3,dum4,a_ar11,a_ar12,a_ar13

            nmu_=0
            dum1=brent(-2.0d0, a_ar1, +2.0d0, ls, tol, a_ar11, t, x, n)
            dum2=brent( a_ar1, 0.5d0*(a_ar1+1.0d0), +2.0d0, ls, tol, a_ar12, t, x, n)
            dum3=brent(-2.0d0, 0.5d0*(a_ar1-1.0d0),  a_ar1, ls, tol, a_ar13, t, x, n)
            if  ((dabs(a_ar12-a_ar11).gt.tol2.and.dabs(a_ar12-a_ar1).gt.tol2) &
            .or.(dabs(a_ar13-a_ar11).gt.tol2.and.dabs(a_ar13-a_ar1).gt.tol2)) &
            nmu_=1
            dum4=dmin1(dum1,dum2,dum3)
            if (dum4.eq.dum2) then
               amin=a_ar12
            else if (dum4.eq.dum3) then
               amin=a_ar13
            else
               amin=a_ar11
            end if
            return
        end subroutine minls

        !-----------------------------
        ! Mean and variance estimation
        !-----------------------------

        subroutine mv(x,n,ave,var)
            use nrtype
            implicit none
            integer, intent(in) :: n
            real(sp), dimension(n), intent(in) :: x
            real(sp), intent(out) :: ave,var
            real(sp), dimension(n) :: y
            real(sp), parameter :: one=1.0_sp
            real(sp), parameter :: two=2.0_sp
            ave=sum(x(:))/n
            y(:)=x(:)-ave
            var=dot_product(y,y)
            var=(var-sum(y)**two/n)/(n-one)
        end subroutine mv

        !-------------------------------------------------
        ! Obtain the number of observations of a data file
        !-------------------------------------------------

        integer function nraws()
            implicit none
            real(sp) :: junk
            integer :: Nobs, i, ios
            Nobs=0
            do i=1,Nobsmax+1
                read (1,*,iostat=ios) junk
                if (ios /=0) exit
                if (i .gt. Nobsmax) then
                    print '(a,i6,a)', ' MAXIMUM NUMBER OF OBSERVATIONS (',Nobsmax,') EXCEEDED'
                    print *
                    print '(a)',      ' Program terminates.'
                    stop
                end if
                Nobs=Nobs+1
                nraws=Nobs
            end do
            if (i .lt. Nobsmin) then
                print '(a,i6,a)', ' MINIMUM NUMBER OF OBSERVATIONS (',Nobsmax,') MISSED'
                print *
                print '(a)',      ' Program terminates.'
                stop
            end if
        end function nraws

        !-------------------
        ! PICKANDS ESTIMATOR
        !-------------------

        subroutine pickands(Y,N,alpha_sequence,N4)
            use nrtype
            implicit none
            integer, intent(in) :: N                                        ! input data size
            real(sp), dimension(N), intent(in) :: Y                         ! input data
            integer, intent(in) :: N4                                       ! output data size
            real(sp), dimension(N4), intent(out) :: alpha_sequence          ! output data
!
!           Pickands estimator of tail index.
!           If pickg = 'y', then tail index = gamma, if pickg = 'n', then tail index = alpha.
!           See Resnick (2007 Heavy-Tail Phenomena: Probabilistic and Statistical Modeling, Springer, New York: pp. 93, 368 therein).
!
            real(sp), allocatable, dimension(:) :: Yordered                 ! auxiliary array
                                                                            ! declaring Yordered allocatable saves RAM
            real(sp), parameter :: one=1.0_sp
            real(sp), parameter :: two=2.0_sp
            integer :: i=-999
            allocate(Yordered(N))
            Yordered=Y
            call sort(Yordered,Y,N,1,N)
            Yordered(1:N)=Yordered(N:1:-1)
            alpha_sequence=(one/log(two))*log((Yordered(1:N4)-Yordered(2:2*N4:2))/(Yordered(2:2*N4:2)-Yordered(4:4*N4:4)))
            if (pickg .eq. 'y') then
               alpha_sequence=alpha_sequence        ! i.e., gamma (this line written for the sake of clarity)
            else
               alpha_sequence=one/alpha_sequence    ! i.e., alpha
            end if
            deallocate(Yordered)
        end subroutine pickands

        !------------------------------------------------------------------------------------------
        !
        ! Data files, input, preprocessing and output
        !
        ! Notes:
        !
        ! (1) Input data Format:   t(1)        Y(1)
        !                          t(2)        Y(2)
        !                          .           .
        !                          .           .
        !                          .           .
        !                          t(Nobs)     Y(Nobs)
        !
        ! (2) Maximum(Nobs) given in function nraws
        !
        ! (3) t(i), Y(i):          o ASCII
        !                          o decimal point allowed
        !                          o exponential notation allowed
        !                          o thousands-comma not allowed
        !
        ! (4) Times t(i) should increase strictly monotonically (but see point (5))
        !
        ! (5) Preprocessing:    o t(i) may be reversed prior to the estimation
        !                       o t(i) and Y(i) may be transformed prior to the estimation
        !
        ! (6) Output: Together with the estimation result and setting, the output file contains:
        !                       o original time series
        !                         (note that notation X and X_original is used instead of
        !                         Y and Y_original)
        !                       o transformed time series
        !
        !     We recommend to write down the performed transformations during the analysis.
        !
        !------------------------------------------------------------------------------------------

        subroutine read_file(t,Y,Nobs,torig,Yorig,Norig,inputfile)
            use cfgseedparameters, only: modus,SUBR_READ_FILE_inputfile
            implicit none
            real(sp), dimension(:), allocatable, intent(inout) :: t     ! time (possibly preprocessed)
            real(sp), dimension(:), allocatable, intent(inout) :: Y     ! variable (possibly preprocessed)
            real(sp), dimension(:), allocatable, intent(inout) :: torig ! time (original, for output)
            real(sp), dimension(:), allocatable, intent(inout) :: Yorig ! variable (original, for output)
            integer, intent(out) :: Nobs
            integer, intent(out) :: Norig
            character*1000, intent(out) :: inputfile                    ! input data file name
            integer :: i                                                ! counter
            if (modus .eq. 'per-hand') then
               print *, 'Enter the name of the data file:'
               read (*,'(A)') inputfile
            else if (modus .eq. 'estimate') then
               inputfile=trim(adjustl(SUBR_READ_FILE_inputfile))
            end if
            open (unit=1, file=inputfile, status='old', form='formatted', action='read')
            Norig=nraws()
            Nobs=Norig
            rewind(1)
            if (allocated(t)) deallocate(t)
            allocate(t(Nobs))
            if (allocated(Y)) deallocate(Y)
            allocate(Y(Nobs))
            if (allocated(torig)) deallocate(torig)
            allocate(torig(Nobs))
            if (allocated(Yorig)) deallocate(Yorig)
            allocate(Yorig(Nobs))
            do i=1,Nobs
               read (1, *) t(i),Y(i)
            end do
            print '(a,i6,a)', ' The file contains ',Nobs,' observations'
            print *
            print '(a)', ' PRESS ENTER TO CONTINUE'
            if (modus .eq. 'per-hand') read (*,*)
            close (1)
            torig=t
            Yorig=Y
        end subroutine read_file

        !------------------------------------------------------
        ! Autocorrelation estimation for persistence estimation
        !------------------------------------------------------

        subroutine rhoest(n,x,rho)
            use nrtype
            implicit none
            integer,  intent(in) :: n
            real(sp), dimension(1:n), intent(in) :: x
            real(sp), intent(out) :: rho
            integer :: i
            real(sp) :: sum1,sum2
            sum1=0.0_sp
            sum2=0.0_sp
            do i=2,n
               sum1=sum1+x(i)*x(i-1)
               sum2=sum2+x(i)**2.0_sp
            end do
            rho=sum1/sum2
        end subroutine rhoest

        !-----------------
        ! SORTING AN ARRAY
        !-----------------

        recursive subroutine sort(Yordered,Y,N,l,r)
            implicit none
!
!           Sorts in ascending order.
!
!           External variables
            integer, intent(in)                 :: N
            integer, intent(in)                 :: l,r
            real(sp), dimension(N), intent(in)  :: Y
            real(sp), dimension(N), intent(inout) :: Yordered
            integer :: i,j
            real(sp) :: v,t1,t2
            i=l
            j=r
            !
            ! Selecting pivot. According to Sedgewick the median of the first, middle, and last value is selected.
            ! Note: this subroutine corresponds to the algorithm Quicksort (sort.f90 in Numerical Recipes).
            !
            v=Yordered(l)
            t1=Yordered(int((l+r)/2))
            t2=Yordered(r)
            if (v<t1) then
                if (t2>t1) then
                    v=t1
                else if (t2>v) then
                    v=t2
                end if
            else
                if (t2<t1) then
                    v=t1
                else if (t2<v) then
                    v=t2
                end if
            end if
            ! Doing partition
            do
                 do while (Yordered(i)<v)
                     i=i+1
                 end do
                 do while(v<Yordered(j))
                     j=j-1
                 end do
                 if (i<=j) then
                     t1=Yordered(i)
                     Yordered(i)=Yordered(j)
                     Yordered(j)=t1
                     i=i+1
                     j=j-1
                 end if
                 if (i>j) exit
            end do
            if (l<j) then
                call sort(Yordered,Y,N,l,j)
            end if
            if (i<r) then
                call sort(Yordered,Y,N,i,r)
            end if
        end subroutine sort

        !--------------------
        ! Stable distribution
        !--------------------

        subroutine stable(alpha,x,n,rng3,rng4)
        use nrtype
        use rng
        implicit none
        real(sp), intent(in)  :: alpha                              ! tail index
        integer, intent(in)  :: n                                   ! data size
        real(sp), dimension(n), intent(out) :: x                    ! variable
        type(rng_t), dimension(1), intent(inout) :: rng3            ! seeded random number generator
        type(rng_t), dimension(1), intent(inout) :: rng4            ! seeded random number generator
!
!       Stable distribution using a modified algorithm after Nolan (1977).
!       Nolan JP 1997 Commun Stat Stoch Models 13:759.
!
!       Details: Modification of previous algorithm (Witt and Malamud 2013 Surv Geophys 34:541),
!       who use the R function rstable, which goes back to Nolan (1997).
!
!       Note: it is ensured that subroutine stable is called with 2 > alpha > 0.
!
!       See also Nolan (2003) for more details about stable distributions.
!       Nolan JP (2003) Modeling financial data with stable distributions. In: Rachev ST (Ed.) Handbook of
!                       Heavy Tailed Distributions in Finance. Elsevier, Amsterdam, 106-130.
!
!       See also Chambers et al. (1976) for an early Fortran implementation of a stable distribution generator.
!       Chambers JM, Mallows CL, Stuck BW (1976) A method for simulating stable random variables. J Am Stat Assoc 71:340.
!
        real(sp), dimension(:), allocatable :: w                    ! auxiliary variable
        integer :: i                                                ! counter
        integer :: iseed=1                                          ! dummy integer for random number generator
        allocate(w(n))
        do i=1,n
           w(i)=mzran(rng3(iseed))
           x(i)=mzran(rng4(iseed))
        end do
        w=pi*(w-0.5_sp)
        x=-log(x)
!
!       Modification of the algorithm by Nolan (1997)
!       to solve numerical problems of type 'NaN' when a large tail index is used.
!
        where (((cos(w-(alpha*w))/x)<0) .and. (((1-alpha)/alpha)<0) .and. (((1-alpha)/alpha)>-1))
              x=(sin(alpha*w)/(cos(w))**(1/alpha))*(abs(cos(w-(alpha*w))/x))**((1-alpha)/alpha)
        elsewhere
              x=(sin(alpha*w)/(cos(w))**(1/alpha))*(cos(w-(alpha*w))/x)**((1-alpha)/alpha)
        end where
        deallocate(w)
        end subroutine stable

        !----------------------------------------
        ! Persistence estimation (uneven spacing)
        !----------------------------------------

        subroutine tauest(t,x,n,tau,rhoout)
            use nrtype
            implicit none
            integer, intent(in) :: n
            real(sp), dimension(n), intent(in)  :: t
            real(sp), dimension(n), intent(in)  :: x
            real(sp),               intent(out) :: tau
            real(sp),               intent(out) :: rhoout
    !
    !       Estimates persistence time, tau, and
    !       equivalent autocorrelation coefficient, rho,
    !       with bias correction for unknown mean
    !       (Mudelsee (2002) Comput Geosci 28:69).
    !
            real(sp), dimension(n) :: tscal, xscal
            real(sp) :: fac, avg, var, dt, rho, scalt, amin, rhoavg
            integer :: i, mult
            real(sp), parameter :: rhoavgmax=0.99_sp        ! maximum
            real(sp), parameter :: rhoavgmin=0.01_sp        ! minimum
            real(sp), parameter :: zero = 0.0_sp
    !
    !       Dummy variables
    !       ===============
    !
            tscal=t
            xscal=x
    !
    !       Centering and scaling of x
    !       ==========================
    !
            call mv(xscal(1:n), n, avg, var)
            xscal(1:n)=xscal(1:n)-avg
            fac = sqrt(var)
            xscal(1:n) = xscal(1:n) / fac
    !
    !       Scaling of t (=> start value of a = 1/e)
    !       ========================================
    !
            dt = (tscal(n)-tscal(1)) / (1.0_sp*(n-1))
            call rhoest(n, xscal(1:n), rho)
            if (rho .le. 0.0_sp) then
                rho = 0.05_sp
            else if (rho .gt. 1.0_sp) then
                rho = 0.95_sp
            end if
    !
    !       This rho setting equal to 0.05 or 0.95 is not problematic since
    !       we require rho only for t scaling.
    !
            scalt = -log(rho)/dt
            tscal(1:n) = tscal(1:n) * scalt
    !
    !       Estimation
    !       ==========
    !
            call minls(n, tscal(1:n), xscal(1:n), amin, mult)
    !
    !       Determine tau
    !       =============
    !
            if (amin .ge. rhoavgmax) then
               rhoavg=rhoavgmax
               tau = -dt /log(rhoavgmax)
            else if (amin .ge. rhoavgmin .and. amin .lt. rhoavgmax) then
               tau = -1.0_sp /(scalt*log(amin))
               rhoavg = exp(-dt / tau)
               rhoavg = (rhoavg * (n-1.0_sp) + 1.0_sp) / (n-4.0_sp) ! Bias correction (unknown mean)
               rhoavg = min(rhoavgmax,rhoavg)
               rhoavg = max(rhoavgmin,rhoavg)
               tau=-dt/log(rhoavg)
            else if (amin .lt. rhoavgmin) then
               rhoavg=rhoavgmin
               tau = zero
            end if
            rhoout=rhoavg
        end subroutine tauest

        !-------------------------------
        ! Test for alpha between 0 and 2
        !-------------------------------

        subroutine test_alpha_initial(Y,Nobs,opt2,flagtest)
            use nrtype
            use Hill_Pick_testNpos_alpha, only: fracp,minNpositive
            implicit none
            integer, intent(in)                    :: Nobs
            real(sp), dimension(Nobs), intent(in)  :: Y         ! variable (possibly preprocessed)
            integer, intent(in)                    :: opt2      ! opt2 = 1: Hill
                                                                ! opt2 = 2: Pickands
            character(len=1), intent(out) :: flagtest           ! flagtest = '+': test passed
                                                                ! flagtest = 'i': test failed
!
!           Test (over the full order search range) whether at least for one order value
!           the tail index alpha is between 0 and 2.
!
!           This test is performed initially to exclude the situation where none of the order search values
!           produces an alpha between 0 and 2.
!
!           If the order selection is based on generating data from a stable distribution with alpha between 0 and 2
!           (see subroutines est_out and stable), then such a situtation would mean that no order selection,
!           and hence no alpha estimation, is possible.
!
            real(sp), dimension(:), allocatable :: Y_est                ! generated process based on the original Y
            real(sp), dimension(:), allocatable :: alpha_sequence       ! sequence of tail index estimates for all possible k values
            integer :: K                                                ! K - 1 = maximum possible search value for the order
            integer :: i                                                ! counter
            real(sp), parameter :: zero=0.0_sp
            real(sp), parameter :: two=2.0_sp
!
! 1.        Preliminary
!           ===========
!
            allocate(Y_est(Nobs))
            flagtest='-'
!
! 2.        Subtract mean
!           =============
!
            Y_est=Y-sum(Y)/Nobs
!
! 3.        Calculate the maximum possible search value for the order
!           =========================================================
!
            if (opt2 .eq. 1) then       ! HILL ESTIMATOR
                K=count(Y_est > zero)
            else if (opt2 .eq. 2) then  ! PICKANDS ESTIMATOR
                K=int(Nobs*fracp)+1
            end if
            if (K .lt. minNpositive) then
               flagtest='-'
               deallocate(Y_est)
               return
            end if
            allocate(alpha_sequence(K-1))
            alpha_sequence=-999.0_sp
!
! 4.        Calculate the alpha sequence
!           ============================
!
            if (opt2 .eq. 1) then       ! HILL ESTIMATOR
                call hill(Y_est,Nobs,alpha_sequence,K)
            else if (opt2 .eq. 2) then  ! PICKANDS ESTIMATOR
                call pickands(Y_est,Nobs,alpha_sequence,K-1)
            end if
!
! 5.        Test
!           ====
!
            do i=1,K-1
               if (alpha_sequence(i) .gt. zero .and. alpha_sequence(i) .lt. two) flagtest='+'
            end do
!
! 6.        Final stuff
!           ===========
!
            deallocate(alpha_sequence)
            deallocate(Y_est)
        end subroutine test_alpha_initial

        !---------------------------
        ! Write additional plot data
        !---------------------------
        
        subroutine writeykopt(t,Y,Nobs,k_opt)
            use nrtype
            implicit none
            integer, intent(in) :: Nobs
            real(sp), dimension(Nobs), intent(in) :: t
            real(sp), dimension(Nobs), intent(in) :: Y
            integer, intent(in)  :: k_opt
!           Writes Y(kopt) to file for plotting (output analysis software ht-d-stats).
            character(len=9), parameter :: outfileykopt='ykopt.dat'
            integer :: i
            real(sp), dimension(Nobs) :: Ycopy
            Ycopy=Y
            open (unit=1, file=outfileykopt, status='unknown', form='formatted', action='write')
            call sort(Ycopy,Ycopy,Nobs,1,Nobs)
            Ycopy(1:Nobs)=Ycopy(Nobs:1:-1)
            do i=1,Nobs
               write (unit=1, fmt='(2(es14.6e3,1x))') t(i),Ycopy(k_opt)
            end do
            close (unit=1, status='keep')
        end subroutine writeykopt
            
    end program menu